/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "3dmath/Matrix.h"
#include "convergence/ChooseClassState.h"
#include "convergence/SharedEvents.h"
#include "convergence/ShellStateGlobals.h"
#include "shell/Shell.h"
#include "shell/ShellKeyCodes.h"
#include <stdlib.h>
#include <string.h>

ChooseClassState * ChooseClassState_create(GameSession * gameSession) {
	stemobject_create_implementation(ChooseClassState, init, gameSession)
}

void ChooseClassState_init(compat_type(ChooseClassState *) selfPtr, GameSession * gameSession) {
	ChooseClassState * self = selfPtr;
	
	StemObject_init(self);
	self->dispose = ChooseClassState_dispose;
	self->activate = ChooseClassState_activate;
	self->deactivate = ChooseClassState_deactivate;
	self->draw = ChooseClassState_draw;
	
	self->gameSession = gameSession;
	self->selectedItemIndex = 0;
}

void ChooseClassState_dispose(compat_type(ChooseClassState *) selfPtr) {
	ChooseClassState * self = selfPtr;
	
	StemObject_dispose(self);
}

static void selectClass(ChooseClassState * self, enum PlayerClass class) {
	self->gameSession->playerClass = class;
	if (self->gameSession->preferences->getBoolean(self->gameSession->preferences, "seen_instructions")) {
		self->stateController->transition(self->stateController, "start_game");
		
	} else {
		self->gameSession->showingInstructionsBeforeGame = true;
		self->stateController->transition(self->stateController, "instructions");
	}
}

static void selectItem(ChooseClassState * self, int itemIndex) {
	switch (itemIndex) {
		case 0:
			selectClass(self, PLAYER_CLASS_BERSERKER);
			break;
			
		case 1:
			selectClass(self, PLAYER_CLASS_RANGER);
			break;
			
		case 2:
			selectClass(self, PLAYER_CLASS_SPY);
			break;
			
		case 3:
			selectClass(self, PLAYER_CLASS_DUELIST);
			break;
			
		case 4:
			selectClass(self, PLAYER_CLASS_SPEEDSTER);
			break;
			
		case 5:
			self->stateController->transition(self->stateController, "back");
			break;
	}
}

static bool mouseDown(void * sender, Atom eventID, void * eventData, void * context) {
	ChooseClassState * self = context;
	struct mouseEvent * event = eventData;
	float localY;
	
	if (event->button == 0) {
		localY = (g_windowHeight - event->y) / g_windowHeight * 2.0f - 1.0f;
		if (localY > 0.625f) {
			selectItem(self, 0);
		} else if (localY > 0.375f) {
			selectItem(self, 1);
		} else if (localY > 0.125f) {
			selectItem(self, 2);
		} else if (localY > -0.125f) {
			selectItem(self, 3);
		} else if (localY > -0.375f) {
			selectItem(self, 4);
		} else {
			selectItem(self, 5);
		}
	}
	return true;
}

static bool keyDown(void * sender, Atom eventID, void * eventData, void * context) {
	ChooseClassState * self = context;
	struct keyEvent * event = eventData;
	
	switch (event->keyCode) {
		case KEYBOARD_UP_ARROW:
			self->selectedItemIndex = (self->selectedItemIndex + 5) % 6;
			Shell_redisplay();
			return true;
			
		case KEYBOARD_DOWN_ARROW:
			self->selectedItemIndex = (self->selectedItemIndex + 1) % 6;
			Shell_redisplay();
			return true;
			
		case KEYBOARD_SPACEBAR:
		case KEYBOARD_RETURN_OR_ENTER:
			selectItem(self, self->selectedItemIndex);
			return true;
			
		case KEYBOARD_ESCAPE:
			self->stateController->transition(self->stateController, "back");
			return true;
			
		default:
			return false;
	}
}

void ChooseClassState_activate(compat_type(ChooseClassState *) selfPtr) {
	ChooseClassState * self = selfPtr;
	
	self->selectedItemIndex = self->gameSession->playerClass;
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_MOUSE_DOWN), mouseDown, self);
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_DOWN), keyDown, self);
	self->font = self->gameSession->resourceManager->referenceResource(self->gameSession->resourceManager, "bitmapfont", "font.json");
	Shell_redisplay();
}

void ChooseClassState_deactivate(compat_type(ChooseClassState *) selfPtr) {
	ChooseClassState * self = selfPtr;
	
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_MOUSE_DOWN), mouseDown, self);
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_DOWN), keyDown, self);
	self->gameSession->resourceManager->releaseResource(self->gameSession->resourceManager, "bitmapfont", "font.json");
}

static void drawString(ChooseClassState * self, bool highlighted, const char * string, float y, float textHeight) {
	float stringWidth;
	
	if (highlighted) {
		glColor4f(1.0f, 0.5f, 0.0f, 1.0f);
	} else {
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	}
	stringWidth = self->font->measureString(self->font, string, strlen(string));
	self->font->drawString(self->font, string, strlen(string), textHeight, -stringWidth * textHeight * 0.5f, y - textHeight * 0.5f, 0.0f);
}

void ChooseClassState_draw(compat_type(ChooseClassState *) selfPtr) {
	ChooseClassState * self = selfPtr;
	float ratio;
	Matrix modelviewMatrix, projectionMatrix;
	
	glClear(GL_COLOR_BUFFER_BIT);
	ratio = (float) g_windowWidth / g_windowHeight;
	projectionMatrix = Matrix_ortho(Matrix_identity(), -ratio, ratio, -1.0f, 1.0f, -1.0f, 1.0f);
	modelviewMatrix = Matrix_identity();
	glMatrixMode(GL_PROJECTION);
	glLoadMatrixf(projectionMatrix.m);
	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(modelviewMatrix.m);
	
	drawString(self, self->selectedItemIndex == 0, "Berserker", 0.75f, 0.09375f);
	drawString(self, self->selectedItemIndex == 0, "Simple melee attack, easy to play", 0.625f + 0.03125f, 0.046875f);
	drawString(self, self->selectedItemIndex == 1, "Ranger", 0.5f, 0.09375f);
	drawString(self, self->selectedItemIndex == 1, "Shoot from a distance, lock with special action", 0.375f + 0.03125f, 0.046875f);
	drawString(self, self->selectedItemIndex == 2, "Spy", 0.25f, 0.09375f);
	drawString(self, self->selectedItemIndex == 2, "Stealth for extra damage with a sneak attack", 0.125f + 0.03125f, 0.046875f);
	drawString(self, self->selectedItemIndex == 3, "Duelist", 0.0f, 0.09375f);
	drawString(self, self->selectedItemIndex == 3, "Block to absorb damage and counterattack", -0.125f + 0.03125f, 0.046875f);
	drawString(self, self->selectedItemIndex == 4, "Speedster", -0.25f, 0.09375f);
	drawString(self, self->selectedItemIndex == 4, "Dash into enemies to attack them", -0.375f + 0.03125f, 0.046875f);
	drawString(self, self->selectedItemIndex == 5, "Back", -0.75f, 0.125f);
}
