/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/CollisionManager.h"
#include "convergence/SharedDefinitions.h"
#include <math.h>
#include <stdio.h>

#define COLLISION_EPSILON (1.0f / 4096.0f) // 0.0001220703125f
#define TIMESLICE_EPSILON (1.0f / 8192.0f) // 0.00006103515625f
#define MAX_TIMESLICE_SUBDIVISIONS 128

static bool intersectSweptAABBStatic(struct sweptAABBGeometry sweptAABB, float staticLeft, float staticRight, float staticBottom, float staticTop, struct collision * outCollision) {
	float timeX = 0.0f, timeY = 0.0f;
	int collisionX = 0, collisionY = 0;
	float newX = 0.0f, newY = 0.0f;
	float movingLeft, movingRight, movingBottom, movingTop;
	float lastMovingLeft, lastMovingRight, lastMovingBottom, lastMovingTop;
	
	movingLeft = sweptAABB.future.x;
	movingRight = movingLeft + sweptAABB.width;
	movingBottom = sweptAABB.future.y;
	movingTop = movingBottom + sweptAABB.height;
	lastMovingLeft = sweptAABB.current.x;
	lastMovingRight = lastMovingLeft + sweptAABB.width;
	lastMovingBottom = sweptAABB.current.y;
	lastMovingTop = lastMovingBottom + sweptAABB.height;
	
	// Incorrect sweeping, but better than nothing
	if (min(movingBottom, lastMovingBottom) < staticTop - COLLISION_EPSILON && max(movingTop, lastMovingTop) > staticBottom + COLLISION_EPSILON) {
		if (lastMovingRight <= staticLeft + COLLISION_EPSILON && movingRight > staticLeft + COLLISION_EPSILON) {
			collisionX = -1;
			// May be wrong after first collision for some objects, since collided object will appear to be moving slower than it actually is (or others appear to be moving faster)?
			timeX = (staticLeft - lastMovingRight) / (movingLeft - lastMovingLeft);
			
		} else if (lastMovingLeft >= staticRight - COLLISION_EPSILON && movingLeft < staticRight - COLLISION_EPSILON) {
			collisionX = 1;
			timeX = (staticRight - lastMovingLeft) / (movingLeft - lastMovingLeft);
		}
		
		// Compensate somewhat for incorrect sweeping by discarding collisions that don't intersect at collision time
		// Without this, collisions with cracks where you're moving perpendicular to the wall faster than parallel to it will be registered
		if (collisionX) {
			newY = lastMovingBottom + (movingBottom - lastMovingBottom) * timeX;
			if (newY >= staticTop - COLLISION_EPSILON || newY + (staticTop - staticBottom) <= staticBottom + COLLISION_EPSILON) {
				collisionX = 0;
			}
		}
	}
	
	if (min(movingLeft, lastMovingLeft) < staticRight - COLLISION_EPSILON && max(movingRight, lastMovingRight) > staticLeft + COLLISION_EPSILON) {
		if (lastMovingTop <= staticBottom + COLLISION_EPSILON && movingTop > staticBottom + COLLISION_EPSILON) {
			collisionY = -1;
			timeY = (staticBottom - lastMovingTop) / (movingBottom - lastMovingBottom);
			
		} else if (lastMovingBottom >= staticTop - COLLISION_EPSILON && movingBottom < staticTop - COLLISION_EPSILON) {
			collisionY = 1;
			timeY = (staticTop - lastMovingBottom) / (movingBottom - lastMovingBottom);
		}
		
		if (collisionY) {
			newX = lastMovingLeft + (movingLeft - lastMovingLeft) * timeY;
			if (newX >= staticRight - COLLISION_EPSILON || newX + (staticRight - staticLeft) <= staticLeft + COLLISION_EPSILON) {
				collisionY = 0;
			}
		}
	}
	
	if (collisionX || collisionY) {
		if (!collisionX || (collisionY && timeY < timeX)) {
			outCollision->time = timeY;
			outCollision->normal.x = 0.0f;
			outCollision->normal.y = collisionY;
			outCollision->surfaceArea = min(staticRight, newX + sweptAABB.width) - max(staticLeft, newX);
			return true;
		}
		
		outCollision->time = timeX;
		outCollision->normal.x = collisionX;
		outCollision->normal.y = 0.0f;
		outCollision->surfaceArea = min(staticTop, newY + sweptAABB.height) - max(staticBottom, newY);
		return true;
	}
	
	return false;
}

static bool intersectLineStatic(struct lineGeometry line, float staticLeft, float staticRight, float staticBottom, float staticTop, struct collision * outCollision) {
	return false;
}

static bool intersectSweptAABBSweptAABB(struct sweptAABBGeometry sweptAABB1, struct sweptAABBGeometry sweptAABB2, struct collision * outCollision) {
	float timeX = 0.0f, timeY = 0.0f;
	int collisionX = 0, collisionY = 0;
	float newX, newY;
	float left1, right1, bottom1, top1;
	float lastLeft1, lastRight1, lastBottom1, lastTop1;
	float left2, right2, bottom2, top2;
	float lastLeft2, lastRight2, lastBottom2, lastTop2;
	
	left1 = sweptAABB1.future.x;
	right1 = left1 + sweptAABB1.width;
	bottom1 = sweptAABB1.future.y;
	top1 = bottom1 + sweptAABB1.height;
	lastLeft1 = sweptAABB1.current.x;
	lastRight1 = lastLeft1 + sweptAABB1.width;
	lastBottom1 = sweptAABB1.current.y;
	lastTop1 = lastBottom1 + sweptAABB1.height;
	left2 = sweptAABB2.future.x;
	right2 = left2 + sweptAABB2.width;
	bottom2 = sweptAABB2.future.y;
	top2 = bottom2 + sweptAABB2.height;
	lastLeft2 = sweptAABB2.current.x;
	lastRight2 = lastLeft2 + sweptAABB2.width;
	lastBottom2 = sweptAABB2.current.y;
	lastTop2 = lastBottom2 + sweptAABB2.height;
	
	// Incorrect sweeping, but better than nothing
	if (min(bottom1, lastBottom1) < max(top2, lastTop2) - COLLISION_EPSILON && max(top1, lastTop1) > min(bottom2, lastBottom2) + COLLISION_EPSILON) {
		if (lastRight1 <= lastLeft2 + COLLISION_EPSILON && right1 > left2 + COLLISION_EPSILON) {
			collisionX = 1;
			timeX = (lastLeft2 - lastRight1) / (left1 - lastLeft1 + lastLeft2 - left2);
			
		} else if (lastLeft1 >= lastRight2 - COLLISION_EPSILON && left1 < right2 - COLLISION_EPSILON) {
			collisionX = -1;
			timeX = (lastRight2 - lastLeft1) / (left1 - lastLeft1 + lastLeft2 - left2);
		}
		
		// TODO: Add sweep compensation test
	}
	
	if (min(left1, lastLeft1) < max(right2, lastRight2) - COLLISION_EPSILON && max(right1, lastRight1) > min(left2, lastLeft2) + COLLISION_EPSILON) {
		if (lastTop1 <= lastBottom2 + COLLISION_EPSILON && top1 > bottom2 + COLLISION_EPSILON) {
			collisionY = 1;
			timeY = (lastBottom2 - lastTop1) / (bottom1 - lastBottom1 + lastBottom2 - bottom2);
			
		} else if (lastBottom1 >= lastTop2 - COLLISION_EPSILON && bottom1 < top2 - COLLISION_EPSILON) {
			collisionY = -1;
			timeY = (lastTop2 - lastBottom1) / (bottom1 - lastBottom1 + lastBottom2 - bottom2);
		}
		
		// TODO: Add sweep compensation test
	}
	
	if (collisionX || collisionY) {
		if (!collisionX || (collisionY && timeY < timeX)) {
			outCollision->time = timeY;
			outCollision->normal.x = 0.0f;
			outCollision->normal.y = collisionY;
			// THIS IS WRONG (but doesn't matter much yet)
			newX = lastLeft1 + (left1 - lastLeft1) * timeY;
			outCollision->surfaceArea = min(right2, newX + sweptAABB1.width) - max(left2, newX);
			return true;
		}
		
		outCollision->time = timeX;
		outCollision->normal.x = collisionX;
		outCollision->normal.y = 0.0f;
		// THIS IS WRONG (but doesn't matter much yet)
		newY = lastBottom1 + (bottom1 - lastBottom1) * timeX;
		outCollision->surfaceArea = min(top2, newY + sweptAABB1.height) - max(bottom2, newY);
		return true;
	}
	
	return false;
}

// If both endpoints of each line are on opposite sides of the other line, they intersect
static bool intersectLineSegments(Vector2 point0, Vector2 point1, Vector2 point2, Vector2 point3) {
	return (Vector2_cross(Vector2_subtract(point1, point0), Vector2_subtract(point2, point0)) > 0.0f) !=
	       (Vector2_cross(Vector2_subtract(point1, point0), Vector2_subtract(point3, point0)) > 0.0f) &&
	       (Vector2_cross(Vector2_subtract(point3, point2), Vector2_subtract(point0, point2)) > 0.0f) !=
	       (Vector2_cross(Vector2_subtract(point3, point2), Vector2_subtract(point1, point2)) > 0.0f);
}

static bool intersectSweptAABBLine(struct sweptAABBGeometry sweptAABB, struct lineGeometry line, struct collision * outCollision) {
	Vector2 aabbVertices[8];
	
	aabbVertices[0] = sweptAABB.current;
	aabbVertices[1].x = sweptAABB.current.x + sweptAABB.width;
	aabbVertices[1].y = sweptAABB.current.y;
	aabbVertices[2].x = sweptAABB.current.x;
	aabbVertices[2].y = sweptAABB.current.y + sweptAABB.height;
	aabbVertices[3].x = sweptAABB.current.x + sweptAABB.width;
	aabbVertices[3].y = sweptAABB.current.y + sweptAABB.height;
	aabbVertices[4] = sweptAABB.future;
	aabbVertices[5].x = sweptAABB.future.x + sweptAABB.width;
	aabbVertices[5].y = sweptAABB.future.y;
	aabbVertices[6].x = sweptAABB.future.x;
	aabbVertices[6].y = sweptAABB.future.y + sweptAABB.height;
	aabbVertices[7].x = sweptAABB.future.x + sweptAABB.width;
	aabbVertices[7].y = sweptAABB.future.y + sweptAABB.height;
	
	if (intersectLineSegments(line.endpoints[0], line.endpoints[1], aabbVertices[0], aabbVertices[1]) ||
	    intersectLineSegments(line.endpoints[0], line.endpoints[1], aabbVertices[0], aabbVertices[2]) ||
	    intersectLineSegments(line.endpoints[0], line.endpoints[1], aabbVertices[1], aabbVertices[3]) ||
	    intersectLineSegments(line.endpoints[0], line.endpoints[1], aabbVertices[2], aabbVertices[3]) ||
	    intersectLineSegments(line.endpoints[0], line.endpoints[1], aabbVertices[4], aabbVertices[5]) ||
	    intersectLineSegments(line.endpoints[0], line.endpoints[1], aabbVertices[4], aabbVertices[6]) ||
	    intersectLineSegments(line.endpoints[0], line.endpoints[1], aabbVertices[5], aabbVertices[7]) ||
	    intersectLineSegments(line.endpoints[0], line.endpoints[1], aabbVertices[6], aabbVertices[7]) ||
	    intersectLineSegments(line.endpoints[0], line.endpoints[1], aabbVertices[0], aabbVertices[4]) ||
	    intersectLineSegments(line.endpoints[0], line.endpoints[1], aabbVertices[1], aabbVertices[5]) ||
	    intersectLineSegments(line.endpoints[0], line.endpoints[1], aabbVertices[2], aabbVertices[6]) ||
	    intersectLineSegments(line.endpoints[0], line.endpoints[1], aabbVertices[3], aabbVertices[7])) {
		outCollision->time = 0.0f;
		outCollision->surfaceArea = 0.0f;
		outCollision->normal = line.effectiveNormal;
		return true;
	}
	return false;
}

static bool intersectLineLine(struct lineGeometry line1, struct lineGeometry line2, struct collision * outCollision) {
	return false;
}

static bool intersectStatic(struct collisionGeometry geometry, float staticLeft, float staticRight, float staticBottom, float staticTop, struct collision * outCollision) {
	switch (geometry.type) {
		case COLLISION_GEOMETRY_TYPE_SWEPT_AABB:
			return intersectSweptAABBStatic(geometry.u.sweptAABB, staticLeft, staticRight, staticBottom, staticTop, outCollision);
			
		case COLLISION_GEOMETRY_TYPE_LINE:
			return intersectLineStatic(geometry.u.line, staticLeft, staticRight, staticBottom, staticTop, outCollision);
	}
	return false;
}

static bool intersect(struct collisionGeometry geometry1, struct collisionGeometry geometry2, struct collision * outCollision) {
	switch (geometry1.type) {
		case COLLISION_GEOMETRY_TYPE_SWEPT_AABB:
			switch (geometry2.type) {
				case COLLISION_GEOMETRY_TYPE_SWEPT_AABB:
					return intersectSweptAABBSweptAABB(geometry1.u.sweptAABB, geometry2.u.sweptAABB, outCollision);
					
				case COLLISION_GEOMETRY_TYPE_LINE:
					return intersectSweptAABBLine(geometry1.u.sweptAABB, geometry2.u.line, outCollision);
			}
			break;
			
		case COLLISION_GEOMETRY_TYPE_LINE:
			switch (geometry2.type) {
				case COLLISION_GEOMETRY_TYPE_SWEPT_AABB:
					if (intersectSweptAABBLine(geometry2.u.sweptAABB, geometry1.u.line, outCollision)) {
						outCollision->normal.x = -outCollision->normal.x;
						outCollision->normal.y = -outCollision->normal.y;
						return true;
					}
					return false;
					
				case COLLISION_GEOMETRY_TYPE_LINE:
					return intersectLineLine(geometry1.u.line, geometry2.u.line, outCollision);
			}
			break;
	}
	return false;
}

void CollisionManager_run(DungeonRoomStateModel * stateModel) {
	GameObjectModel * object, * object2;
	size_t objectIndex, objectIndex2;
	struct rect4f objectRect, tileRect, object2Rect;
	unsigned int tileIndexX, tileIndexY;
	struct collisionGeometry objectGeometry, object2Geometry;
	struct collision collision, bestCollision = {0.0f, 0.0f, {0.0f, 0.0f}};
	bool collided;
	GameObjectModel * bestCollisionObject = NULL, * bestCollisionSecondaryObject = NULL;
	bool bestCollisionStatic = false;
	float timesliceRemaining;
	int numberOfTimesSubdivided;
	
	timesliceRemaining = 1.0f;
	numberOfTimesSubdivided = 0;
	
	while (timesliceRemaining > TIMESLICE_EPSILON) {
		bestCollision.time = 1.0f + TIMESLICE_EPSILON;
		collided = false;
		
		for (objectIndex = 0; objectIndex < stateModel->objectCount; objectIndex++) {
			object = stateModel->objects[objectIndex];
			objectRect = object->roughBounds(object);
			objectGeometry = object->collisionGeometry(object);
			
			tileRect.left = floor(objectRect.left);
			tileRect.right = ceil(objectRect.right);
			tileRect.bottom = floor(objectRect.bottom);
			tileRect.top = ceil(objectRect.top);
			
			if (tileRect.left < 0) {
				tileRect.left = 0;
			} else if (tileRect.left > (int) stateModel->tileCountX) {
				tileRect.left = stateModel->tileCountX;
			}
			if (tileRect.right < 0) {
				tileRect.right = 0;
			} else if (tileRect.right > (int) stateModel->tileCountX) {
				tileRect.right = stateModel->tileCountX;
			}
			if (tileRect.bottom < 0) {
				tileRect.bottom = 0;
			} else if (tileRect.bottom > (int) stateModel->tileCountY) {
				tileRect.bottom = stateModel->tileCountY;
			}
			if (tileRect.top < 0) {
				tileRect.top = 0;
			} else if (tileRect.top > (int) stateModel->tileCountY) {
				tileRect.top = stateModel->tileCountY;
			}
			
			for (tileIndexY = (unsigned int) tileRect.bottom; tileIndexY < (unsigned int) tileRect.top; tileIndexY++) {
				for (tileIndexX = (unsigned int) tileRect.left; tileIndexX < (unsigned int) tileRect.right; tileIndexX++) {
					if (stateModel->tiles[tileIndexY * stateModel->tileCountX + tileIndexX].type == DUNGEON_TILE_TYPE_WALL) {
						if (intersectStatic(objectGeometry, tileIndexX, tileIndexX + 1.0f, tileIndexY, tileIndexY + 1.0f, &collision)) {
							if (collision.time < bestCollision.time ||
							    (collision.time < bestCollision.time + TIMESLICE_EPSILON &&
							     collision.surfaceArea > bestCollision.surfaceArea)) {
								collided = true;
								bestCollision = collision;
								bestCollisionObject = object;
								bestCollisionStatic = true;
							}
						}
					}
				}
			}
			
			for (objectIndex2 = objectIndex + 1; objectIndex2 < stateModel->objectCount; objectIndex2++) {
				object2 = stateModel->objects[objectIndex2];
				object2Rect = object2->roughBounds(object2);
				
				if (objectRect.left <= object2Rect.right && objectRect.right >= object2Rect.left && objectRect.bottom <= object2Rect.top && objectRect.top >= object2Rect.bottom) {
					object2Geometry = object2->collisionGeometry(object2);
					if (object->wouldCollide(object, object2) &&
					    object2->wouldCollide(object2, object) &&
					    intersect(objectGeometry, object2Geometry, &collision)) {
						if (collision.time < bestCollision.time ||
						    (collision.time < bestCollision.time + TIMESLICE_EPSILON &&
						     collision.surfaceArea > bestCollision.surfaceArea)) {
							collided = true;
							bestCollision = collision;
							bestCollisionObject = object;
							bestCollisionSecondaryObject = object2;
							bestCollisionStatic = false;
						}
					}
				}
			}
		}
			
		if (!collided) {
			break;
		}
		
		// TODO: It doesn't seem right to pass both bestCollision and timesliceRemaining to collision(), but only bestCollision.time to interpolate()
		
		if (bestCollisionStatic) {
			bestCollisionObject->collision(bestCollisionObject, NULL, bestCollision, timesliceRemaining);
		} else {
			GameObjectModel * object1Copy, * object2Copy;
			
			// Passing copies ensures both order independence and immutability
			object1Copy = bestCollisionObject->copy(bestCollisionObject);
			object2Copy = bestCollisionSecondaryObject->copy(bestCollisionSecondaryObject);
			bestCollisionObject->collision(bestCollisionObject, object2Copy, bestCollision, timesliceRemaining);
			bestCollision.normal.x = -bestCollision.normal.x;
			bestCollision.normal.y = -bestCollision.normal.y;
			bestCollisionSecondaryObject->collision(bestCollisionSecondaryObject, object1Copy, bestCollision, timesliceRemaining);
			object1Copy->dispose(object1Copy);
			object2Copy->dispose(object2Copy);
		}
		
		for (objectIndex = 0; objectIndex < stateModel->objectCount; objectIndex++) {
			object = stateModel->objects[objectIndex];
			if (object == bestCollisionObject || (!bestCollisionStatic && object == bestCollisionSecondaryObject)) {
				continue;
			}
			object->interpolate(object, bestCollision.time);
		}
		
		timesliceRemaining = (1.0f - bestCollision.time) * timesliceRemaining;
		numberOfTimesSubdivided++;
		
		if (numberOfTimesSubdivided > MAX_TIMESLICE_SUBDIVISIONS) {
			fprintf(stderr, "Warning: Maximum timeslice subdivisons reached; collision resolution may be incomplete\n");
			break;
		}
	}
}
