/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "3dmath/Matrix.h"
#include "convergence/AudioManager.h"
#include "convergence/ConfigureControlsState.h"
#include "convergence/SharedEvents.h"
#include "convergence/ShellStateGlobals.h"
#include "gamepad/Gamepad.h"
#include "shell/Shell.h"
#include "shell/ShellKeyCodes.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

ConfigureControlsState * ConfigureControlsState_create(GameSession * gameSession) {
	stemobject_create_implementation(ConfigureControlsState, init, gameSession)
}

void ConfigureControlsState_init(compat_type(ConfigureControlsState *) selfPtr, GameSession * gameSession) {
	ConfigureControlsState * self = selfPtr;
	
	StemObject_init(self);
	self->dispose = ConfigureControlsState_dispose;
	self->activate = ConfigureControlsState_activate;
	self->deactivate = ConfigureControlsState_deactivate;
	self->draw = ConfigureControlsState_draw;
	
	self->gameSession = gameSession;
	self->gamepadViewCount = 0;
	self->gamepadViews = NULL;
	self->lastBindingMessage[127] = 0;
}

void ConfigureControlsState_dispose(compat_type(ConfigureControlsState *) selfPtr) {
	ConfigureControlsState * self = selfPtr;
	
	free(self->gamepadViews);
	StemObject_dispose(self);
}

static void controlClicked(ConfigureControlsState * self, int index) {
	if (index == self->selectedControl) {
		self->selectedControl = -1;
	} else {
		self->selectedControl = index;
	}
	self->lastBindingMessage[0] = 0;
}

static void controlBoundSuccessfully(ConfigureControlsState * self) {
	switch (self->selectedControl) {
		case 0:
		case 1:
		case 2:
		case 3:
			AudioManager_play("dash");
			break;
		case 4:
			AudioManager_play("sword_swing");
			break;
		case 5:
			AudioManager_play("sneak");
			break;
	}
	self->selectedControl = -1;
}

static const char * const actionIDs[] = {
	"left",
	"right",
	"down",
	"up",
	"attack",
	"special"
};

#define TRIGGER_THRESHOLD 0.5f
#define RELEASE_THRESHOLD 0.375f

static bool mouseDown(void * sender, Atom eventID, void * eventData, void * context) {
	ConfigureControlsState * self = context;
	struct mouseEvent * event = eventData;
	float ratio, localX, localY;
	
	if (event->button == 0) {
		ratio = (float) g_windowWidth / (float) g_windowHeight;
		localX = event->x / g_windowHeight * 2.0f - ratio;
		localY = (g_windowHeight - event->y) / g_windowHeight * 2.0f - 1.0f;
		if (localY < -0.675f) {
			self->gameSession->saveControlBindings(self->gameSession);
			self->stateController->transition(self->stateController, "done");
			
		} else if (localY > 0.75f) {
			if (localX < -ratio + ratio / 3.0f * 1) {
				controlClicked(self, 0);
			} else if (localX < -ratio + ratio / 3.0f * 2) {
				controlClicked(self, 1);
			} else if (localX < -ratio + ratio / 3.0f * 3) {
				controlClicked(self, 2);
			} else if (localX < -ratio + ratio / 3.0f * 4) {
				controlClicked(self, 3);
			} else if (localX < -ratio + ratio / 3.0f * 5) {
				controlClicked(self, 4);
			} else {
				controlClicked(self, 5);
			}
			
		} else if (localY < 0.0f && self->selectedControl != -1) {
			size_t gamepadViewIndex;
			float offsetY = 0.0f, scale;
			struct GamepadDetailView_mappableElement element;
			
			for (gamepadViewIndex = 0; gamepadViewIndex < self->gamepadViewCount; gamepadViewIndex++) {
				scale = ratio * 2.0f / self->gamepadViews[gamepadViewIndex]->width;
				if (localY <= offsetY && localY >= offsetY - 0.03125f * scale) {
					element = self->gamepadViews[gamepadViewIndex]->mappableElementAtPosition(self->gamepadViews[gamepadViewIndex], (localX + ratio) / scale);
					if (element.axis) {
						if (self->gameSession->playerInputMap->isAxisBound(self->gameSession->playerInputMap,
						                                                   Atom_fromString(actionIDs[self->selectedControl]),
						                                                   self->gamepadViews[gamepadViewIndex]->device->vendorID,
						                                                   self->gamepadViews[gamepadViewIndex]->device->productID,
						                                                   element.u.axis.axisID,
						                                                   element.u.axis.direction)) {
							self->gameSession->playerInputMap->unbindAxis(self->gameSession->playerInputMap,
							                                              Atom_fromString(actionIDs[self->selectedControl]),
							                                              self->gamepadViews[gamepadViewIndex]->device->vendorID,
							                                              self->gamepadViews[gamepadViewIndex]->device->productID,
							                                              element.u.axis.axisID,
							                                              element.u.axis.direction);
						} else {
							self->gameSession->playerInputMap->bindAxis(self->gameSession->playerInputMap,
							                                            Atom_fromString(actionIDs[self->selectedControl]),
							                                            self->gamepadViews[gamepadViewIndex]->device->vendorID,
							                                            self->gamepadViews[gamepadViewIndex]->device->productID,
							                                            element.u.axis.axisID,
							                                            element.u.axis.direction,
							                                            TRIGGER_THRESHOLD * element.u.axis.direction,
							                                            RELEASE_THRESHOLD * element.u.axis.direction);
						}
					} else {
						if (self->gameSession->playerInputMap->isButtonBound(self->gameSession->playerInputMap,
						                                                    Atom_fromString(actionIDs[self->selectedControl]),
						                                                    self->gamepadViews[gamepadViewIndex]->device->vendorID,
						                                                    self->gamepadViews[gamepadViewIndex]->device->productID,
						                                                    element.u.button.buttonID)) {
							self->gameSession->playerInputMap->unbindButton(self->gameSession->playerInputMap,
							                                                Atom_fromString(actionIDs[self->selectedControl]),
							                                                self->gamepadViews[gamepadViewIndex]->device->vendorID,
							                                                self->gamepadViews[gamepadViewIndex]->device->productID,
							                                                element.u.button.buttonID);
						} else {
							self->gameSession->playerInputMap->bindButton(self->gameSession->playerInputMap,
							                                              Atom_fromString(actionIDs[self->selectedControl]),
							                                              self->gamepadViews[gamepadViewIndex]->device->vendorID,
							                                              self->gamepadViews[gamepadViewIndex]->device->productID,
							                                              element.u.button.buttonID);
						}
					}
					
					controlBoundSuccessfully(self);
					break;
				}
				offsetY -= 0.03125f * scale + 0.03125f;
			}
		}
	}
	return true;
}

static bool keyDown(void * sender, Atom eventID, void * eventData, void * context) {
	struct keyEvent * event = eventData;
	ConfigureControlsState * self = context;
	
	if (event->keyCode == KEYBOARD_RETURN_OR_ENTER || event->keyCode == KEYBOARD_ESCAPE) {
		self->stateController->transition(self->stateController, "done");
		return true;
	}
	
	if (self->selectedControl != -1) {
		if (self->gameSession->playerInputMap->isKeyBound(self->gameSession->playerInputMap,
		                                                  Atom_fromString(actionIDs[self->selectedControl]),
		                                                  event->keyCode)) {
			self->gameSession->playerInputMap->unbindKey(self->gameSession->playerInputMap,
			                                             Atom_fromString(actionIDs[self->selectedControl]),
			                                             event->keyCode);
		} else {
			self->gameSession->playerInputMap->bindKey(self->gameSession->playerInputMap,
			                                           Atom_fromString(actionIDs[self->selectedControl]),
			                                           event->charCode,
			                                           event->keyCode);
		}
		controlBoundSuccessfully(self);
	}
	return self->inputController->keyDown(self->inputController, event->charCode, event->keyCode, event->modifiers);
}

static bool keyUp(void * sender, Atom eventID, void * eventData, void * context) {
	struct keyEvent * event = eventData;
	ConfigureControlsState * self = context;
	
	return self->inputController->keyUp(self->inputController, event->charCode, event->keyCode, event->modifiers);
}

static bool keyModifiersChanged(void * sender, Atom eventID, void * eventData, void * context) {
	struct keyEvent * event = eventData;
	ConfigureControlsState * self = context;
	
	return self->inputController->keyModifiersChanged(self->inputController, event->modifiers);
}

static bool deviceAttached(void * sender, Atom eventID, void * eventData, void * context) {
	ConfigureControlsState * self = context;
	struct Gamepad_device * device = eventData;
	
	self->gamepadViews = realloc(self->gamepadViews, sizeof(GamepadDetailView *) * (self->gamepadViewCount + 1));
	self->gamepadViews[self->gamepadViewCount++] = GamepadDetailView_create(device);
	
	return true;
}

static bool deviceRemoved(void * sender, Atom eventID, void * eventData, void * context) {
	ConfigureControlsState * self = context;
	struct Gamepad_device * device = eventData;
	size_t gamepadViewIndex;
	
	for (gamepadViewIndex = 0; gamepadViewIndex < self->gamepadViewCount; gamepadViewIndex++) {
		if (self->gamepadViews[gamepadViewIndex]->device == device) {
			self->gamepadViews[gamepadViewIndex]->dispose(self->gamepadViews[gamepadViewIndex]);
			self->gamepadViewCount--;
			for (; gamepadViewIndex < self->gamepadViewCount; gamepadViewIndex++) {
				self->gamepadViews[gamepadViewIndex] = self->gamepadViews[gamepadViewIndex + 1];
			}
			break;
		}
	}
	
	return true;
}

static const char * controlIndexToString(int controlIndex) {
	switch (controlIndex) {
		case 0:
			return "Left";
		case 1:
			return "Right";
		case 2:
			return "Down";
		case 3:
			return "Up";
		case 4:
			return "Attack";
		case 5:
			return "Special";
	}
	return "Unknown";
}

static const char * shellKeyCodeToString(unsigned int keyCode) {
	switch (keyCode) {
		case KEYBOARD_ESCAPE:
			return "Escape";
		case KEYBOARD_F1:
			return "F1";
		case KEYBOARD_F2:
			return "F2";
		case KEYBOARD_F3:
			return "F3";
		case KEYBOARD_F4:
			return "F4";
		case KEYBOARD_F5:
			return "F5";
		case KEYBOARD_F6:
			return "F6";
		case KEYBOARD_F7:
			return "F7";
		case KEYBOARD_F8:
			return "F8";
		case KEYBOARD_F9:
			return "F9";
		case KEYBOARD_F10:
			return "F10";
		case KEYBOARD_F11:
			return "F11";
		case KEYBOARD_F12:
			return "F12";
		case KEYBOARD_F13:
			return "F13";
		case KEYBOARD_F14:
			return "F14";
		case KEYBOARD_F15:
			return "F15";
		case KEYBOARD_GRAVE_ACCENT_AND_TILDE:
			return "~";
		case KEYBOARD_1:
			return "1";
		case KEYBOARD_2:
			return "2";
		case KEYBOARD_3:
			return "3";
		case KEYBOARD_4:
			return "4";
		case KEYBOARD_5:
			return "5";
		case KEYBOARD_6:
			return "6";
		case KEYBOARD_7:
			return "7";
		case KEYBOARD_8:
			return "8";
		case KEYBOARD_9:
			return "9";
		case KEYBOARD_0:
			return "0";
		case KEYBOARD_HYPHEN:
			return "-";
		case KEYBOARD_EQUAL_SIGN:
			return "=";
		case KEYBOARD_DELETE_OR_BACKSPACE:
			return "Delete";
		case KEYBOARD_TAB:
			return "Tab";
		case KEYBOARD_Q:
			return "Q";
		case KEYBOARD_W:
			return "W";
		case KEYBOARD_E:
			return "E";
		case KEYBOARD_R:
			return "R";
		case KEYBOARD_T:
			return "T";
		case KEYBOARD_Y:
			return "Y";
		case KEYBOARD_U:
			return "U";
		case KEYBOARD_I:
			return "I";
		case KEYBOARD_O:
			return "O";
		case KEYBOARD_P:
			return "P";
		case KEYBOARD_OPEN_BRACKET:
			return "[";
		case KEYBOARD_CLOSE_BRACKET:
			return "]";
		case KEYBOARD_BACKSLASH:
			return "\\";
		case KEYBOARD_A:
			return "A";
		case KEYBOARD_S:
			return "S";
		case KEYBOARD_D:
			return "D";
		case KEYBOARD_F:
			return "F";
		case KEYBOARD_G:
			return "G";
		case KEYBOARD_H:
			return "H";
		case KEYBOARD_J:
			return "J";
		case KEYBOARD_K:
			return "K";
		case KEYBOARD_L:
			return "L";
		case KEYBOARD_SEMICOLON:
			return ";";
		case KEYBOARD_QUOTE:
			return "\"";
		case KEYBOARD_RETURN_OR_ENTER:
			return "Enter";
		case KEYBOARD_Z:
			return "Z";
		case KEYBOARD_X:
			return "X";
		case KEYBOARD_C:
			return "C";
		case KEYBOARD_V:
			return "V";
		case KEYBOARD_B:
			return "B";
		case KEYBOARD_N:
			return "N";
		case KEYBOARD_M:
			return "M";
		case KEYBOARD_COMMA:
			return ",";
		case KEYBOARD_PERIOD:
			return ".";
		case KEYBOARD_SLASH:
			return "/";
		case KEYBOARD_SPACEBAR:
			return "Spacebar";
		case KEYBOARD_LEFT_ARROW:
			return "Left arrow";
		case KEYBOARD_UP_ARROW:
			return "Up arrow";
		case KEYBOARD_DOWN_ARROW:
			return "Down arrow";
		case KEYBOARD_RIGHT_ARROW:
			return "Right arrow";
		case KEYBOARD_INSERT:
			return "Insert";
		case KEYBOARD_HOME:
			return "Home";
		case KEYBOARD_PAGE_UP:
			return "Page up";
		case KEYBOARD_DELETE_FORWARD:
			return "Forward delete";
		case KEYBOARD_END:
			return "End";
		case KEYBOARD_PAGE_DOWN:
			return "Page down";
		case KEYPAD_NUM_LOCK:
			return "Num lock";
		case KEYPAD_EQUAL_SIGN:
			return "Keypad =";
		case KEYPAD_SLASH:
			return "Keypad /";
		case KEYPAD_ASTERISK:
			return "Keypad *";
		case KEYPAD_HYPHEN:
			return "Keypad -";
		case KEYPAD_PLUS:
			return "Keypad +";
		case KEYPAD_PERIOD:
			return "Keypad .";
		case KEYPAD_ENTER:
			return "Keypad enter";
		case KEYPAD_1:
			return "Keypad 1";
		case KEYPAD_2:
			return "Keypad 2";
		case KEYPAD_3:
			return "Keypad 3";
		case KEYPAD_4:
			return "Keypad 4";
		case KEYPAD_5:
			return "Keypad 5";
		case KEYPAD_6:
			return "Keypad 6";
		case KEYPAD_7:
			return "Keypad 7";
		case KEYPAD_8:
			return "Keypad 8";
		case KEYPAD_9:
			return "Keypad 9";
		case KEYPAD_0:
			return "Keypad 0";
	}
	return "unknown key";
}

static int indexOfGamepad(int vendorID, int productID) {
	unsigned int gamepadIndex;
	struct Gamepad_device * device;
	
	for (gamepadIndex = 0; gamepadIndex < Gamepad_numDevices(); gamepadIndex++) {
		device = Gamepad_deviceAtIndex(gamepadIndex);
		if (device->vendorID == vendorID && device->productID == productID) {
			return gamepadIndex;
		}
	}
	return -1;
}

static bool keyBound(void * sender, Atom eventID, void * eventData, void * context) {
	ConfigureControlsState * self = context;
	struct PlayerInputMap_keyboardBinding * event = eventData;
	
	snprintf(self->lastBindingMessage, 127, "%s bound to %s", controlIndexToString(self->selectedControl), shellKeyCodeToString(event->keyCode));
	return true;
}

static bool keyUnbound(void * sender, Atom eventID, void * eventData, void * context) {
	ConfigureControlsState * self = context;
	struct PlayerInputMap_keyboardBinding * event = eventData;
	
	snprintf(self->lastBindingMessage, 127, "%s unbound from %s", controlIndexToString(self->selectedControl), shellKeyCodeToString(event->keyCode));
	return true;
}

static bool buttonBound(void * sender, Atom eventID, void * eventData, void * context) {
	ConfigureControlsState * self = context;
	struct PlayerInputMap_buttonBindingEvent * event = eventData;
	
	snprintf(self->lastBindingMessage, 127, "%s bound to gamepad %d button %d", controlIndexToString(self->selectedControl), indexOfGamepad(event->vendorID, event->productID) + 1, event->binding.buttonID + 1);
	return true;
}

static bool buttonUnbound(void * sender, Atom eventID, void * eventData, void * context) {
	ConfigureControlsState * self = context;
	struct PlayerInputMap_buttonBindingEvent * event = eventData;
	
	snprintf(self->lastBindingMessage, 127, "%s unbound from gamepad %d button %d", controlIndexToString(self->selectedControl), indexOfGamepad(event->vendorID, event->productID) + 1, event->binding.buttonID + 1);
	return true;
}

static bool axisBound(void * sender, Atom eventID, void * eventData, void * context) {
	ConfigureControlsState * self = context;
	struct PlayerInputMap_axisBindingEvent * event = eventData;
	
	snprintf(self->lastBindingMessage, 127, "%s bound to gamepad %d axis %d %s", controlIndexToString(self->selectedControl), indexOfGamepad(event->vendorID, event->productID) + 1, event->binding.axisID + 1, event->binding.direction == -1 ? "negative" : "positive");
	return true;
}

static bool axisUnbound(void * sender, Atom eventID, void * eventData, void * context) {
	ConfigureControlsState * self = context;
	struct PlayerInputMap_axisBindingEvent * event = eventData;
	
	snprintf(self->lastBindingMessage, 127, "%s unbound from gamepad %d axis %d %s", controlIndexToString(self->selectedControl), indexOfGamepad(event->vendorID, event->productID) + 1, event->binding.axisID + 1, event->binding.direction == -1 ? "negative" : "positive");
	return true;
}

static bool backgrounded(void * sender, Atom eventID, void * eventData, void * context) {
	ConfigureControlsState * self = context;
	
	self->backgrounded = true;
	return true;
}

static bool foregrounded(void * sender, Atom eventID, void * eventData, void * context) {
	ConfigureControlsState * self = context;
	
	self->backgrounded = false;
	Shell_redisplay();
	return true;
}

void ConfigureControlsState_activate(compat_type(ConfigureControlsState *) selfPtr) {
	ConfigureControlsState * self = selfPtr;
	unsigned int deviceIndex;
	
	self->inputController = InputController_create(self->gameSession->playerInputMap);
	self->selectedControl = -1;
	self->backgrounded = false;
	self->font = self->gameSession->resourceManager->referenceResource(self->gameSession->resourceManager, "bitmapfont", "font.json");
	self->lastBindingMessage[0] = 0;
	
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_MOUSE_DOWN), mouseDown, self);
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_DOWN), keyDown, self);
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_UP), keyUp, self);
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_MODIFIERS_CHANGED), keyModifiersChanged, self);
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_BACKGROUNDED), backgrounded, self);
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_FOREGROUNDED), foregrounded, self);
	
	Gamepad_detectDevices();
	Gamepad_processEvents();
	Gamepad_eventDispatcher()->registerForEvent(Gamepad_eventDispatcher(), Atom_fromString(GAMEPAD_EVENT_DEVICE_ATTACHED), deviceAttached, self);
	Gamepad_eventDispatcher()->registerForEvent(Gamepad_eventDispatcher(), Atom_fromString(GAMEPAD_EVENT_DEVICE_REMOVED), deviceRemoved, self);
	for (deviceIndex = 0; deviceIndex < Gamepad_numDevices(); deviceIndex++) {
		deviceAttached(NULL, GAMEPAD_EVENT_DEVICE_ATTACHED, Gamepad_deviceAtIndex(deviceIndex), self);
	}
	
	self->gameSession->playerInputMap->eventDispatcher->registerForEvent(self->gameSession->playerInputMap->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_KEY_BOUND), keyBound, self);
	self->gameSession->playerInputMap->eventDispatcher->registerForEvent(self->gameSession->playerInputMap->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_KEY_UNBOUND), keyUnbound, self);
	self->gameSession->playerInputMap->eventDispatcher->registerForEvent(self->gameSession->playerInputMap->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_BUTTON_BOUND), buttonBound, self);
	self->gameSession->playerInputMap->eventDispatcher->registerForEvent(self->gameSession->playerInputMap->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_BUTTON_UNBOUND), buttonUnbound, self);
	self->gameSession->playerInputMap->eventDispatcher->registerForEvent(self->gameSession->playerInputMap->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_AXIS_BOUND), axisBound, self);
	self->gameSession->playerInputMap->eventDispatcher->registerForEvent(self->gameSession->playerInputMap->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_AXIS_UNBOUND), axisUnbound, self);
	
	Shell_redisplay();
}

void ConfigureControlsState_deactivate(compat_type(ConfigureControlsState *) selfPtr) {
	ConfigureControlsState * self = selfPtr;
	unsigned int deviceIndex;
	
	self->inputController->dispose(self->inputController);
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_MOUSE_DOWN), mouseDown, self);
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_DOWN), keyDown, self);
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_UP), keyUp, self);
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_MODIFIERS_CHANGED), keyModifiersChanged, self);
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_BACKGROUNDED), backgrounded, self);
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_FOREGROUNDED), foregrounded, self);
	self->gameSession->resourceManager->releaseResource(self->gameSession->resourceManager, "bitmapfont", "font.json");
	Gamepad_eventDispatcher()->unregisterForEvent(Gamepad_eventDispatcher(), Atom_fromString(GAMEPAD_EVENT_DEVICE_ATTACHED), deviceAttached, self);
	Gamepad_eventDispatcher()->unregisterForEvent(Gamepad_eventDispatcher(), Atom_fromString(GAMEPAD_EVENT_DEVICE_REMOVED), deviceRemoved, self);
	for (deviceIndex = 0; deviceIndex < Gamepad_numDevices(); deviceIndex++) {
		deviceRemoved(NULL, GAMEPAD_EVENT_DEVICE_REMOVED, Gamepad_deviceAtIndex(deviceIndex), self);
	}
	self->gameSession->playerInputMap->eventDispatcher->unregisterForEvent(self->gameSession->playerInputMap->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_KEY_BOUND), keyBound, self);
	self->gameSession->playerInputMap->eventDispatcher->unregisterForEvent(self->gameSession->playerInputMap->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_KEY_UNBOUND), keyUnbound, self);
	self->gameSession->playerInputMap->eventDispatcher->unregisterForEvent(self->gameSession->playerInputMap->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_BUTTON_BOUND), buttonBound, self);
	self->gameSession->playerInputMap->eventDispatcher->unregisterForEvent(self->gameSession->playerInputMap->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_BUTTON_UNBOUND), buttonUnbound, self);
	self->gameSession->playerInputMap->eventDispatcher->unregisterForEvent(self->gameSession->playerInputMap->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_AXIS_BOUND), axisBound, self);
	self->gameSession->playerInputMap->eventDispatcher->unregisterForEvent(self->gameSession->playerInputMap->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_AXIS_UNBOUND), axisUnbound, self);
}

#define CONTROL_HEIGHT 0.046875f

static void drawControl(ConfigureControlsState * self, int index, const char * label, bool triggered, float ratio) {
	float stringWidth;
	
	if (index == self->selectedControl) {
		if (triggered) {
			glColor4f(1.0f, 1.0f, 0.0f, 1.0f);
		} else {
			glColor4f(0.0f, 1.0f, 0.0f, 1.0f);
		}
	} else {
		if (triggered) {
			glColor4f(1.0f, 0.5f, 0.0f, 1.0f);
		} else {
			glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		}
	}
	stringWidth = self->font->measureString(self->font, label, strlen(label));
	self->font->drawString(self->font, label, strlen(label), CONTROL_HEIGHT, -ratio + ratio / 3.0f * index + ratio / 6.0f - stringWidth * CONTROL_HEIGHT * 0.5f, 0.875f - CONTROL_HEIGHT, 0.0f);
}

void drawString(ConfigureControlsState * self, const char * string, float y, float textHeight) {
	float stringWidth;
	
	stringWidth = self->font->measureString(self->font, string, strlen(string));
	self->font->drawString(self->font, string, strlen(string), textHeight, -stringWidth * textHeight * 0.5f, y - textHeight * 0.5f, 0.0f);
}

void ConfigureControlsState_draw(compat_type(ConfigureControlsState *) selfPtr) {
	ConfigureControlsState * self = selfPtr;
	float ratio;
	Matrix modelviewMatrix, projectionMatrix;
	size_t gamepadViewIndex;
	float scale;
	char string[32];
	
	self->inputController->update(self->inputController);
	
	glClear(GL_COLOR_BUFFER_BIT);
	ratio = (float) g_windowWidth / g_windowHeight;
	projectionMatrix = Matrix_ortho(Matrix_identity(), -ratio, ratio, -1.0f, 1.0f, -1.0f, 1.0f);
	modelviewMatrix = Matrix_identity();
	glMatrixMode(GL_PROJECTION);
	glLoadMatrixf(projectionMatrix.m);
	
	Matrix_translate(&modelviewMatrix, -ratio, 0.0f, 0.0f);
	for (gamepadViewIndex = 0; gamepadViewIndex < self->gamepadViewCount; gamepadViewIndex++) {
		scale = ratio * 2.0f / self->gamepadViews[gamepadViewIndex]->width;
		Matrix_translate(&modelviewMatrix, 0.0f, -0.03125f * scale, 0.0f);
		self->gamepadViews[gamepadViewIndex]->draw(self->gamepadViews[gamepadViewIndex], Matrix_scaled(modelviewMatrix, scale, scale, 1.0f));
		Matrix_translate(&modelviewMatrix, 0.0f, -0.03125f, 0.0f);
	}
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	drawControl(self, 0, "[ Left ]", self->inputController->leftTriggered, ratio);
	drawControl(self, 1, "[ Right]", self->inputController->rightTriggered, ratio);
	drawControl(self, 2, "[ Down ]", self->inputController->downTriggered, ratio);
	drawControl(self, 3, "[  Up  ]", self->inputController->upTriggered, ratio);
	drawControl(self, 4, "[Attack]", self->inputController->attackTriggered, ratio);
	drawControl(self, 5, "[Special]", self->inputController->specialTriggered, ratio);
	
	glColor4f(0.5f, 1.0f, 0.25f, 1.0f);
	drawString(self, self->lastBindingMessage, 0.65625f, 0.046875f);
	
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	if (self->selectedControl == -1) {
		drawString(self, "Click an action above to bind it.", 0.5f, 0.0625f);
		drawString(self, "Bound controls highlight when pressed.", 0.375f, 0.046875f);
		if (Gamepad_numDevices() == 0) {
			drawString(self, "Connect a USB gamepad to configure it.", 0.0f, 0.0625f);
		}
	} else {
		sprintf(string, "Press a key to bind to %s.", controlIndexToString(self->selectedControl));
		drawString(self, string, 0.5f, 0.0625);
		if (Gamepad_numDevices() == 0) {
			drawString(self, "Connect a USB gamepad to configure it.", 0.0f, 0.0625f);
		} else {
			drawString(self, "Click a gamepad button or axis to bind to it.", 0.375f, 0.046875f);
			drawString(self, "When binding an axis, click left or right side.", 0.25f + 0.03125, 0.046875f);
		}
	}
	drawString(self, "Done", -0.75f, 0.125f);
	
	if (!self->backgrounded) {
		Shell_redisplay();
	}
}
