/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "3dmath/Matrix.h"
#include "convergence/InstructionsState.h"
#include "convergence/SharedEvents.h"
#include "convergence/ShellStateGlobals.h"
#include "shell/Shell.h"
#include "shell/ShellKeyCodes.h"
#include <stdlib.h>
#include <string.h>

InstructionsState * InstructionsState_create(GameSession * gameSession) {
	stemobject_create_implementation(InstructionsState, init, gameSession)
}

void InstructionsState_init(compat_type(InstructionsState *) selfPtr, GameSession * gameSession) {
	InstructionsState * self = selfPtr;
	
	StemObject_init(self);
	self->dispose = InstructionsState_dispose;
	self->activate = InstructionsState_activate;
	self->deactivate = InstructionsState_deactivate;
	self->draw = InstructionsState_draw;
	
	self->gameSession = gameSession;
}

void InstructionsState_dispose(compat_type(InstructionsState *) selfPtr) {
	InstructionsState * self = selfPtr;
	
	StemObject_dispose(self);
}

static void exitScreen(InstructionsState * self) {
	self->gameSession->preferences->setBoolean(self->gameSession->preferences, "seen_instructions", true);
	self->gameSession->preferences->save(self->gameSession->preferences);
	self->stateController->transition(self->stateController, self->gameSession->showingInstructionsBeforeGame ? "start_game" : "back");
}

static bool mouseDown(void * sender, Atom eventID, void * eventData, void * context) {
	InstructionsState * self = context;
	struct mouseEvent * event = eventData;
	float localY;
	
	if (event->button == 0) {
		localY = (g_windowHeight - event->y) / g_windowHeight * 2.0f - 1.0f;
		if (localY <= -0.625f) {
			exitScreen(self);
		}
	}
	return true;
}

static bool keyDown(void * sender, Atom eventID, void * eventData, void * context) {
	InstructionsState * self = context;
	struct keyEvent * event = eventData;
	
	switch (event->keyCode) {
		case KEYBOARD_SPACEBAR:
		case KEYBOARD_RETURN_OR_ENTER:
			exitScreen(self);
			return true;
			
		case KEYBOARD_ESCAPE:
			if (!self->gameSession->showingInstructionsBeforeGame) {
				exitScreen(self);
				return true;
			}
			break;
	}
	return false;
}

void InstructionsState_activate(compat_type(InstructionsState *) selfPtr) {
	InstructionsState * self = selfPtr;
	
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_MOUSE_DOWN), mouseDown, self);
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_DOWN), keyDown, self);
	self->font = self->gameSession->resourceManager->referenceResource(self->gameSession->resourceManager, "bitmapfont", "font.json");
	Shell_redisplay();
}

void InstructionsState_deactivate(compat_type(InstructionsState *) selfPtr) {
	InstructionsState * self = selfPtr;
	
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_MOUSE_DOWN), mouseDown, self);
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_DOWN), keyDown, self);
	self->gameSession->resourceManager->releaseResource(self->gameSession->resourceManager, "bitmapfont", "font.json");
}

static void drawString(InstructionsState * self, bool highlighted, const char * string, float y, float textHeight) {
	float stringWidth;
	
	if (highlighted) {
		glColor4f(1.0f, 0.5f, 0.0f, 1.0f);
	} else {
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	}
	stringWidth = self->font->measureString(self->font, string, strlen(string));
	self->font->drawString(self->font, string, strlen(string), textHeight, -stringWidth * textHeight * 0.5f, y - textHeight * 0.5f, 0.0f);
}

void InstructionsState_draw(compat_type(InstructionsState *) selfPtr) {
	InstructionsState * self = selfPtr;
	float ratio;
	Matrix modelviewMatrix, projectionMatrix;
	
	glClear(GL_COLOR_BUFFER_BIT);
	ratio = (float) g_windowWidth / g_windowHeight;
	projectionMatrix = Matrix_ortho(Matrix_identity(), -ratio, ratio, -1.0f, 1.0f, -1.0f, 1.0f);
	modelviewMatrix = Matrix_identity();
	glMatrixMode(GL_PROJECTION);
	glLoadMatrixf(projectionMatrix.m);
	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(modelviewMatrix.m);
	
	drawString(self, false, "INSTRUCTIONS", 0.75f, 0.125);
	drawString(self, false, "Move with arrow keys", 0.375f, 0.0625);
	drawString(self, false, "Attack with Z", 0.25f, 0.0625);
	drawString(self, false, "Use special ability with X", 0.125f, 0.0625);
	drawString(self, false, "(spy, ranger, and duelist only)", 0.0f, 0.0625);
	drawString(self, false, "Pick up gems to open doors", -0.125f, 0.0625);
	drawString(self, false, "Pick up apples to heal yourself", -0.25f, 0.0625);
	drawString(self, false, "Get to the end of the dungeon to win", -0.375f, 0.0625);
	drawString(self, true, self->gameSession->showingInstructionsBeforeGame ? "Play" : "Back", -0.75f, 0.125f);
}
