/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/MovingEntityModel.h"
#include <stdlib.h>

#define FACING_INTENT_DECAY_TIME 10

MovingEntityModel * MovingEntityModel_create(DungeonRoomStateModel * stateModel) {
	stemobject_create_implementation(MovingEntityModel, init, stateModel)
}

static void sharedInit(MovingEntityModel * self) {
	self->dispose = MovingEntityModel_dispose;
	self->copy = MovingEntityModel_copy;
	self->update = MovingEntityModel_update;
	self->atRest = MovingEntityModel_atRest;
	self->roughBounds = MovingEntityModel_roughBounds;
	self->collisionGeometry = MovingEntityModel_collisionGeometry;
	self->wouldCollide = MovingEntityModel_wouldCollide;
	self->collision = MovingEntityModel_collision;
	self->interpolate = MovingEntityModel_interpolate;
	self->protected_ivar(collisionInternal) = MovingEntityModel_collisionInternal;
	self->startMovingLeft = MovingEntityModel_startMovingLeft;
	self->startMovingRight = MovingEntityModel_startMovingRight;
	self->startMovingDown = MovingEntityModel_startMovingDown;
	self->startMovingUp = MovingEntityModel_startMovingUp;
	self->stopMovingLeft = MovingEntityModel_stopMovingLeft;
	self->stopMovingRight = MovingEntityModel_stopMovingRight;
	self->stopMovingDown = MovingEntityModel_stopMovingDown;
	self->stopMovingUp = MovingEntityModel_stopMovingUp;
	self->stopMoving = MovingEntityModel_stopMoving;
	
	self->radius = 0.4375f;
	self->acceleration = 0.001f;
	self->deceleration = 0.002f;
	self->velocityMax = 0.01f;
}

void MovingEntityModel_init(compat_type(MovingEntityModel *) selfPtr, DungeonRoomStateModel * stateModel) {
	MovingEntityModel * self = selfPtr;
	
	GameObjectModel_init(self, stateModel);
	sharedInit(self);
	
	self->lastPosition.x = 0.0f;
	self->lastPosition.y = 0.0f;
	self->position.x = 0.0f;
	self->position.y = 0.0f;
	self->facing = NORTH;
	self->velocity.x = 0.0f;
	self->velocity.y = 0.0f;
	self->moveIntentX = 0;
	self->moveIntentY = 0;
	self->facingIntentX = 0;
	self->facingIntentY = 0;
	self->facingIntentXDecay = 0;
	self->facingIntentYDecay = 0;
}

void MovingEntityModel_initCopy(compat_type(MovingEntityModel *) selfPtr, compat_type(MovingEntityModel *) originalPtr) {
	MovingEntityModel * self = selfPtr, * original = originalPtr;
	
	GameObjectModel_initCopy(self, original);
	sharedInit(self);
	
	self->lastPosition = original->lastPosition;
	self->position = original->position;
	self->facing = original->facing;
	self->velocity = original->velocity;
	self->moveIntentX = original->moveIntentX;
	self->moveIntentY = original->moveIntentY;
	self->facingIntentX = original->facingIntentX;
	self->facingIntentY = original->facingIntentY;
	self->facingIntentXDecay = original->facingIntentXDecay;
	self->facingIntentYDecay = original->facingIntentYDecay;
}

void MovingEntityModel_dispose(compat_type(MovingEntityModel *) selfPtr) {
	GameObjectModel_dispose(selfPtr);
}

compat_type(MovingEntityModel *) MovingEntityModel_copy(compat_type(MovingEntityModel *) selfPtr) {
	stemobject_copy_implementation(MovingEntityModel, initCopy)
}

void MovingEntityModel_update(compat_type(MovingEntityModel *) selfPtr) {
	MovingEntityModel * self = selfPtr;
	Vector2 deceleration = {0.0f, 0.0f};
	
	if (self->moveIntentX == 0 || fabs(self->velocity.x) > self->velocityMax) {
		if (self->velocity.x < 0.0f) {
			deceleration.x = 1.0f;
		} else if (self->velocity.x > 0.0f) {
			deceleration.x = -1.0f;
		}
		
		if (self->facingIntentXDecay > 0) {
			self->facingIntentXDecay--;
			if (self->facingIntentXDecay == 0 && self->moveIntentY != 0) {
				self->facingIntentX = 0;
			}
		}
		
	} else {
		if (self->velocity.x * self->moveIntentX < 0) {
			self->velocity.x += self->deceleration * self->moveIntentX;
		} else {
			self->velocity.x += self->acceleration * self->moveIntentX;
		}
		if (self->velocity.x < -self->velocityMax) {
			self->velocity.x = -self->velocityMax;
		} else if (self->velocity.x > self->velocityMax) {
			self->velocity.x = self->velocityMax;
		}
	}
	
	if (self->moveIntentY == 0 || fabs(self->velocity.y) > self->velocityMax) {
		if (self->velocity.y < 0.0f) {
			deceleration.y = 1.0f;
		} else if (self->velocity.y > 0.0f) {
			deceleration.y = -1.0f;
		}
		
		if (self->facingIntentYDecay > 0) {
			self->facingIntentYDecay--;
			if (self->facingIntentYDecay == 0 && self->moveIntentX != 0) {
				self->facingIntentY = 0;
			}
		}
		
	} else {
		if (self->velocity.y * self->moveIntentY < 0) {
			self->velocity.y += self->deceleration * self->moveIntentY;
		} else {
			self->velocity.y += self->acceleration * self->moveIntentY;
		}
		if (self->velocity.y < -self->velocityMax) {
			self->velocity.y = -self->velocityMax;
		} else if (self->velocity.y > self->velocityMax) {
			self->velocity.y = self->velocityMax;
		}
	}
	
	if (deceleration.x != 0.0f || deceleration.y != 0.0f) {
		if (deceleration.x != 0.0f && deceleration.y != 0.0f) {
			deceleration.x = 0.707106781186548f * deceleration.x;
			deceleration.y = 0.707106781186548f * deceleration.y;
		}
		deceleration.x *= self->deceleration;
		deceleration.y *= self->deceleration;
		self->velocity.x += deceleration.x;
		if (deceleration.x < 0.0f && self->velocity.x < 0.0f) {
			self->velocity.x = 0.0f;
		} else if (deceleration.x > 0.0f && self->velocity.x > 0.0f) {
			self->velocity.x = 0.0f;
		}
		self->velocity.y += deceleration.y;
		if (deceleration.y < 0.0f && self->velocity.y < 0.0f) {
			self->velocity.y = 0.0f;
		} else if (deceleration.y > 0.0f && self->velocity.y > 0.0f) {
			self->velocity.y = 0.0f;
		}
	}
	
	if (self->facingIntentX != 0 && self->facingIntentY != 0) {
		self->facing = intentToDirection(self->facingIntentX, self->facingIntentY);
	} else if (self->moveIntentX != 0 || self->moveIntentY != 0) {
		self->facing = intentToDirection(self->moveIntentX, self->moveIntentY);
	}
	
	self->lastPosition = self->position;
	self->position.x += self->velocity.x;
	self->position.y += self->velocity.y;
}

bool MovingEntityModel_atRest(compat_type(MovingEntityModel *) selfPtr) {
	MovingEntityModel * self = selfPtr;
	
	return self->moveIntentX == 0 && self->moveIntentY == 0 && self->velocity.x == 0.0f && self->velocity.y == 0.0f;
}

struct rect4f MovingEntityModel_roughBounds(compat_type(MovingEntityModel *) selfPtr) {
	MovingEntityModel * self = selfPtr;
	struct rect4f bounds;
	
	bounds.left = min(self->position.x, self->lastPosition.x) - self->radius;
	bounds.right = max(self->position.x, self->lastPosition.x) + self->radius;
	bounds.bottom = min(self->position.y, self->lastPosition.y) - self->radius;
	bounds.top = max(self->position.y, self->lastPosition.y) + self->radius;
	
	return bounds;
}

struct collisionGeometry MovingEntityModel_collisionGeometry(compat_type(MovingEntityModel *) selfPtr) {
	MovingEntityModel * self = selfPtr;
	struct collisionGeometry geometry;
	
	geometry.type = COLLISION_GEOMETRY_TYPE_SWEPT_AABB;
	geometry.u.sweptAABB.current.x = self->lastPosition.x - self->radius;
	geometry.u.sweptAABB.current.y = self->lastPosition.y - self->radius;
	geometry.u.sweptAABB.future.x = self->position.x - self->radius;
	geometry.u.sweptAABB.future.y = self->position.y - self->radius;
	geometry.u.sweptAABB.width = self->radius * 2.0f;
	geometry.u.sweptAABB.height = self->radius * 2.0f;
	return geometry;
}

bool MovingEntityModel_wouldCollide(compat_type(MovingEntityModel *) selfPtr, compat_type(GameObjectModel *) objectPtr) {
	switch (((GameObjectModel *) objectPtr)->type) {
		case GAME_OBJECT_TYPE_ENEMY:
		case GAME_OBJECT_TYPE_PLAYER:
		case GAME_OBJECT_TYPE_SWORD_SWING:
		case GAME_OBJECT_TYPE_PROJECTILE:
			return true;
		default:
			return false;
	}
}

void MovingEntityModel_collision(compat_type(MovingEntityModel *) selfPtr, compat_type(GameObjectModel *) objectPtr, const struct collision collision, float time) {
	MovingEntityModel * self = selfPtr;
	
	self->position.x = self->lastPosition.x + (self->position.x - self->lastPosition.x) * collision.time;
	self->position.y = self->lastPosition.y + (self->position.y - self->lastPosition.y) * collision.time;
	
	self->protected_ivar(collisionInternal)(self, objectPtr, collision, time);
	
	self->lastPosition = self->position;
	self->position.x += self->velocity.x * (1.0f - collision.time) * time;
	self->position.y += self->velocity.y * (1.0f - collision.time) * time;
}

void MovingEntityModel_interpolate(compat_type(MovingEntityModel *) selfPtr, float time) {
	MovingEntityModel * self = selfPtr;
	
	self->lastPosition.x += (self->position.x - self->lastPosition.x) * time;
	self->lastPosition.y += (self->position.y - self->lastPosition.y) * time;
}

void MovingEntityModel_collisionInternal(compat_type(MovingEntityModel *) selfPtr, compat_type(GameObjectModel *) objectPtr, const struct collision collision, float time) {
	MovingEntityModel * self = selfPtr;
	GameObjectModel * object = objectPtr;
	float dot;
	
	dot = Vector2_dot(collision.normal, self->velocity);
	if (object == NULL || object->type == GAME_OBJECT_TYPE_ROOM_EXIT) {
		self->velocity.x -= collision.normal.x * dot;
		self->velocity.y -= collision.normal.y * dot;
		
	} else {
		if (object->type == GAME_OBJECT_TYPE_ENEMY || object->type == GAME_OBJECT_TYPE_PLAYER) {
			MovingEntityModel * movingEntity = objectPtr;
			float dot2;
			
			dot2 = Vector2_dot(collision.normal, movingEntity->velocity);
			self->velocity.x -= (collision.normal.x * dot - collision.normal.x * dot2) * 0.5f;
			self->velocity.y -= (collision.normal.y * dot - collision.normal.y * dot2) * 0.5f;
			
		} else if (object->type == GAME_OBJECT_TYPE_DOOR) {
			self->velocity.x -= collision.normal.x * dot;
			self->velocity.y -= collision.normal.y * dot;
		}
	}
}

void MovingEntityModel_startMovingLeft(compat_type(MovingEntityModel *) selfPtr) {
	MovingEntityModel * self = selfPtr;
	
	if (self->moveIntentX != -1) {
		self->moveIntentX = self->facingIntentX = -1;
		self->facingIntentY = self->moveIntentY;
	}
}

void MovingEntityModel_startMovingRight(compat_type(MovingEntityModel *) selfPtr) {
	MovingEntityModel * self = selfPtr;
	
	if (self->moveIntentX != 1) {
		self->moveIntentX = self->facingIntentX = 1;
		self->facingIntentY = self->moveIntentY;
	}
}

void MovingEntityModel_startMovingDown(compat_type(MovingEntityModel *) selfPtr) {
	MovingEntityModel * self = selfPtr;
	
	if (self->moveIntentY != -1) {
		self->moveIntentY = self->facingIntentY = -1;
		self->facingIntentX = self->moveIntentX;
	}
}

void MovingEntityModel_startMovingUp(compat_type(MovingEntityModel *) selfPtr) {
	MovingEntityModel * self = selfPtr;
	
	if (self->moveIntentY != 1) {
		self->moveIntentY = self->facingIntentY = 1;
		self->facingIntentX = self->moveIntentX;
	}
}

void MovingEntityModel_stopMovingLeft(compat_type(MovingEntityModel *) selfPtr) {
	MovingEntityModel * self = selfPtr;
	
	if (self->moveIntentX == -1) {
		self->moveIntentX = 0;
		self->facingIntentXDecay = FACING_INTENT_DECAY_TIME;
	}
}

void MovingEntityModel_stopMovingRight(compat_type(MovingEntityModel *) selfPtr) {
	MovingEntityModel * self = selfPtr;
	
	if (self->moveIntentX == 1) {
		self->moveIntentX = 0;
		self->facingIntentXDecay = FACING_INTENT_DECAY_TIME;
	}
}

void MovingEntityModel_stopMovingDown(compat_type(MovingEntityModel *) selfPtr) {
	MovingEntityModel * self = selfPtr;
	
	if (self->moveIntentY == -1) {
		self->moveIntentY = 0;
		self->facingIntentYDecay = FACING_INTENT_DECAY_TIME;
	}
}

void MovingEntityModel_stopMovingUp(compat_type(MovingEntityModel *) selfPtr) {
	MovingEntityModel * self = selfPtr;
	
	if (self->moveIntentY == 1) {
		self->moveIntentY = 0;
		self->facingIntentYDecay = FACING_INTENT_DECAY_TIME;
	}
}

void MovingEntityModel_stopMoving(compat_type(MovingEntityModel *) selfPtr) {
	MovingEntityModel * self = selfPtr;
	
	self->moveIntentX = 0;
	self->moveIntentY = 0;
	self->facingIntentXDecay = self->facingIntentYDecay = FACING_INTENT_DECAY_TIME;
}
