/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/PlayerInputMap.h"
#include "convergence/SharedDefinitions.h"

PlayerInputMap * PlayerInputMap_create() {
	stemobject_create_implementation(PlayerInputMap, init)
}

void PlayerInputMap_init(compat_type(PlayerInputMap *) selfPtr) {
	PlayerInputMap * self = selfPtr;
	
	StemObject_init(self);
	self->dispose = PlayerInputMap_dispose;
	self->isKeyBound = PlayerInputMap_isKeyBound;
	self->bindKey = PlayerInputMap_bindKey;
	self->unbindKey = PlayerInputMap_unbindKey;
	self->isButtonBound = PlayerInputMap_isButtonBound;
	self->bindButton = PlayerInputMap_bindButton;
	self->unbindButton = PlayerInputMap_unbindButton;
	self->isAxisBound = PlayerInputMap_isAxisBound;
	self->bindAxis = PlayerInputMap_bindAxis;
	self->unbindAxis = PlayerInputMap_unbindAxis;
	
	self->eventDispatcher = EventDispatcher_create(self);
	self->keyboardBindingCount = 0;
	self->keyboardBindings = NULL;
	self->gamepadMapCount = 0;
	self->gamepadMaps = NULL;
}

void PlayerInputMap_dispose(compat_type(PlayerInputMap *) selfPtr) {
	PlayerInputMap * self = selfPtr;
	
	self->eventDispatcher->dispose(self->eventDispatcher);
	StemObject_dispose(self);
}

PlayerInputMap * PlayerInputMap_deserialize(compat_type(DeserializationContext *) deserializationContext) {
	stemobject_deserialize_implementation(PlayerInputMap, deserializationContext, loadSerializedData)
}

bool PlayerInputMap_loadSerializedData(compat_type(PlayerInputMap *) selfPtr, compat_type(DeserializationContext *) deserializationContext) {
	PlayerInputMap * self = selfPtr;
	DeserializationContext * context = deserializationContext;
	uint16_t formatVersion;
	size_t keyboardBindingCount, bindingIndex;
	struct PlayerInputMap_keyboardBinding * keyboardBindings = NULL;
	const char * actionID;
	size_t gamepadMapCount, gamepadMapIndex;
	struct PlayerInputMap_gamepadMap * gamepadMaps;
	
	context->beginStructure(context, "player_input_map");
	formatVersion = context->readUInt16(context, "format_version");
	if (context->status == SERIALIZATION_ERROR_OK && formatVersion != PLAYER_INPUT_MAP_SERIALIZATION_FORMAT_VERSION) {
		return false;
	}
	keyboardBindingCount = context->beginArray(context, "keyboard_bindings");
	keyboardBindings = malloc(sizeof(struct PlayerInputMap_keyboardBinding) * keyboardBindingCount);
	for (bindingIndex = 0; bindingIndex < keyboardBindingCount; bindingIndex++) {
		context->beginStructure(context, NULL);
		keyboardBindings[bindingIndex].actionID = Atom_fromString(context->readString(context, "action"));
		keyboardBindings[bindingIndex].keyCode = context->readUInt32(context, "key_code");
		keyboardBindings[bindingIndex].charCode = context->readUInt32(context, "char_code");
		context->endStructure(context);
	}
	context->endArray(context);
	gamepadMapCount = context->beginArray(context, "gamepad_maps");
	gamepadMaps = calloc(sizeof(struct PlayerInputMap_gamepadMap), gamepadMapCount);
	for (gamepadMapIndex = 0; gamepadMapIndex < gamepadMapCount; gamepadMapIndex++) {
		context->beginStructure(context, NULL);
		gamepadMaps[gamepadMapIndex].vendorID = context->readInt32(context, "vendor_id");
		gamepadMaps[gamepadMapIndex].productID = context->readInt32(context, "product_id");
		gamepadMaps[gamepadMapIndex].buttonBindingCount = context->beginDictionary(context, "button_bindings");
		gamepadMaps[gamepadMapIndex].buttonBindings = malloc(sizeof(struct PlayerInputMap_gamepadButtonBinding) * gamepadMaps[gamepadMapIndex].buttonBindingCount);
		for (bindingIndex = 0; bindingIndex < gamepadMaps[gamepadMapIndex].buttonBindingCount; bindingIndex++) {
			actionID = context->readNextDictionaryKey(context);
			gamepadMaps[gamepadMapIndex].buttonBindings[bindingIndex].actionID = Atom_fromString(actionID);
			gamepadMaps[gamepadMapIndex].buttonBindings[bindingIndex].buttonID = context->readUInt32(context, actionID);
		}
		context->endDictionary(context);
		gamepadMaps[gamepadMapIndex].axisBindingCount = context->beginArray(context, "axis_bindings");
		gamepadMaps[gamepadMapIndex].axisBindings = malloc(sizeof(struct PlayerInputMap_gamepadAxisBinding) * gamepadMaps[gamepadMapIndex].axisBindingCount);
		for (bindingIndex = 0; bindingIndex < gamepadMaps[gamepadMapIndex].axisBindingCount; bindingIndex++) {
			context->beginStructure(context, NULL);
			gamepadMaps[gamepadMapIndex].axisBindings[bindingIndex].actionID = Atom_fromString(context->readString(context, "action"));
			gamepadMaps[gamepadMapIndex].axisBindings[bindingIndex].axisID = context->readUInt32(context, "axis_id");
			gamepadMaps[gamepadMapIndex].axisBindings[bindingIndex].direction = context->readInt8(context, "direction");
			gamepadMaps[gamepadMapIndex].axisBindings[bindingIndex].triggerThreshold = context->readFloat(context, "trigger_threshold");
			gamepadMaps[gamepadMapIndex].axisBindings[bindingIndex].releaseThreshold = context->readFloat(context, "release_threshold");
			context->endStructure(context);
		}
		context->endArray(context);
		context->endStructure(context);
	}
	context->endArray(context);
	context->endStructure(context);
	
	if (context->status != SERIALIZATION_ERROR_OK) {
		free(keyboardBindings);
		for (gamepadMapIndex = 0; gamepadMapIndex < gamepadMapCount; gamepadMapIndex++) {
			free(gamepadMaps[gamepadMapIndex].buttonBindings);
			free(gamepadMaps[gamepadMapIndex].axisBindings);
		}
		free(gamepadMaps);
		return false;
	}
	
	PlayerInputMap_init(self);
	self->keyboardBindingCount = keyboardBindingCount;
	self->keyboardBindings = keyboardBindings;
	self->gamepadMapCount = gamepadMapCount;
	self->gamepadMaps = gamepadMaps;
	
	return true;
}

void PlayerInputMap_serialize(compat_type(PlayerInputMap *) selfPtr, compat_type(SerializationContext *) serializationContext) {
	PlayerInputMap * self = selfPtr;
	SerializationContext * context = serializationContext;
	size_t bindingIndex;
	size_t gamepadMapIndex;
	
	context->beginStructure(context, "player_input_map");
	context->writeUInt16(context, "format_version", PLAYER_INPUT_MAP_SERIALIZATION_FORMAT_VERSION);
	context->beginArray(context, "keyboard_bindings");
	for (bindingIndex = 0; bindingIndex < self->keyboardBindingCount; bindingIndex++) {
		context->beginStructure(context, NULL);
		context->writeString(context, "action", self->keyboardBindings[bindingIndex].actionID);
		context->writeUInt32(context, "key_code", self->keyboardBindings[bindingIndex].keyCode);
		context->writeUInt32(context, "char_code", self->keyboardBindings[bindingIndex].charCode);
		context->endStructure(context);
	}
	context->endArray(context);
	context->beginArray(context, "gamepad_maps");
	for (gamepadMapIndex = 0; gamepadMapIndex < self->gamepadMapCount; gamepadMapIndex++) {
		context->beginStructure(context, NULL);
		context->writeInt32(context, "vendor_id", self->gamepadMaps[gamepadMapIndex].vendorID);
		context->writeInt32(context, "product_id", self->gamepadMaps[gamepadMapIndex].productID);
		context->beginDictionary(context, "button_bindings");
		for (bindingIndex = 0; bindingIndex < self->gamepadMaps[gamepadMapIndex].buttonBindingCount; bindingIndex++) {
			context->writeUInt32(context, self->gamepadMaps[gamepadMapIndex].buttonBindings[bindingIndex].actionID, self->gamepadMaps[gamepadMapIndex].buttonBindings[bindingIndex].buttonID);
		}
		context->endDictionary(context);
		context->beginArray(context, "axis_bindings");
		for (bindingIndex = 0; bindingIndex < self->gamepadMaps[gamepadMapIndex].axisBindingCount; bindingIndex++) {
			context->beginStructure(context, NULL);
			context->writeString(context, "action", self->gamepadMaps[gamepadMapIndex].axisBindings[bindingIndex].actionID);
			context->writeUInt32(context, "axis_id", self->gamepadMaps[gamepadMapIndex].axisBindings[bindingIndex].axisID);
			context->writeInt8(context, "direction", self->gamepadMaps[gamepadMapIndex].axisBindings[bindingIndex].direction);
			context->writeFloat(context, "trigger_threshold", self->gamepadMaps[gamepadMapIndex].axisBindings[bindingIndex].triggerThreshold);
			context->writeFloat(context, "release_threshold", self->gamepadMaps[gamepadMapIndex].axisBindings[bindingIndex].releaseThreshold);
			context->endStructure(context);
		}
		context->endArray(context);
		context->endStructure(context);
	}
	context->endArray(context);
	context->endStructure(context);
}

bool PlayerInputMap_isKeyBound(compat_type(PlayerInputMap *) selfPtr, Atom actionID, unsigned int keyCode) {
	PlayerInputMap * self = selfPtr;
	size_t bindingIndex;
	
	for (bindingIndex = 0; bindingIndex < self->keyboardBindingCount; bindingIndex++) {
		if (self->keyboardBindings[bindingIndex].actionID == actionID && self->keyboardBindings[bindingIndex].keyCode == keyCode) {
			return true;
		}
	}
	return false;
}

void PlayerInputMap_bindKey(compat_type(PlayerInputMap *) selfPtr, Atom actionID, unsigned int charCode, unsigned int keyCode) {
	PlayerInputMap * self = selfPtr;
	size_t bindingIndex;
	
	for (bindingIndex = 0; bindingIndex < self->keyboardBindingCount; bindingIndex++) {
		if (self->keyboardBindings[bindingIndex].actionID == actionID && self->keyboardBindings[bindingIndex].keyCode == keyCode) {
			self->keyboardBindings[bindingIndex].charCode = charCode;
			self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_KEY_BOUND), &self->keyboardBindings[bindingIndex]);
			return;
		}
	}
	self->keyboardBindings = realloc(self->keyboardBindings, sizeof(struct PlayerInputMap_keyboardBinding) * (self->keyboardBindingCount + 1));
	self->keyboardBindings[self->keyboardBindingCount].actionID = actionID;
	self->keyboardBindings[self->keyboardBindingCount].charCode = charCode;
	self->keyboardBindings[self->keyboardBindingCount].keyCode = keyCode;
	self->keyboardBindingCount++;
	
	self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_KEY_BOUND), &self->keyboardBindings[self->keyboardBindingCount - 1]);
}

void PlayerInputMap_unbindKey(compat_type(PlayerInputMap *) selfPtr, Atom actionID, unsigned int keyCode) {
	PlayerInputMap * self = selfPtr;
	size_t bindingIndex;
	struct PlayerInputMap_keyboardBinding binding;
	
	for (bindingIndex = 0; bindingIndex < self->keyboardBindingCount; bindingIndex++) {
		if (self->keyboardBindings[bindingIndex].actionID == actionID && self->keyboardBindings[bindingIndex].keyCode == keyCode) {
			binding = self->keyboardBindings[bindingIndex];
			self->keyboardBindingCount--;
			for (; bindingIndex < self->keyboardBindingCount; bindingIndex++) {
				self->keyboardBindings[bindingIndex] = self->keyboardBindings[bindingIndex + 1];
			}
			self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_KEY_UNBOUND), &binding);
			break;
		}
	}
}

bool PlayerInputMap_isButtonBound(compat_type(PlayerInputMap *) selfPtr, Atom actionID, int vendorID, int productID, unsigned int buttonID) {
	PlayerInputMap * self = selfPtr;
	size_t gamepadIndex, bindingIndex;
	
	for (gamepadIndex = 0; gamepadIndex < self->gamepadMapCount; gamepadIndex++) {
		if (self->gamepadMaps[gamepadIndex].vendorID == vendorID && self->gamepadMaps[gamepadIndex].productID == productID) {
			for (bindingIndex = 0; bindingIndex < self->gamepadMaps[gamepadIndex].buttonBindingCount; bindingIndex++) {
				if (self->gamepadMaps[gamepadIndex].buttonBindings[bindingIndex].actionID == actionID && self->gamepadMaps[gamepadIndex].buttonBindings[bindingIndex].buttonID == buttonID) {
					return true;
				}
			}
		}
	}
	return false;
}

void PlayerInputMap_bindButton(compat_type(PlayerInputMap *) selfPtr, Atom actionID, int vendorID, int productID, unsigned int buttonID) {
	PlayerInputMap * self = selfPtr;
	size_t gamepadIndex, bindingIndex;
	struct PlayerInputMap_buttonBindingEvent event;
	
	for (gamepadIndex = 0; gamepadIndex < self->gamepadMapCount; gamepadIndex++) {
		if (self->gamepadMaps[gamepadIndex].vendorID == vendorID && self->gamepadMaps[gamepadIndex].productID == productID) {
			for (bindingIndex = 0; bindingIndex < self->gamepadMaps[gamepadIndex].buttonBindingCount; bindingIndex++) {
				if (self->gamepadMaps[gamepadIndex].buttonBindings[bindingIndex].actionID == actionID && self->gamepadMaps[gamepadIndex].buttonBindings[bindingIndex].buttonID == buttonID) {
					return;
				}
			}
			self->gamepadMaps[gamepadIndex].buttonBindings = realloc(self->gamepadMaps[gamepadIndex].buttonBindings, sizeof(struct PlayerInputMap_gamepadButtonBinding) * (self->gamepadMaps[gamepadIndex].buttonBindingCount + 1));
			self->gamepadMaps[gamepadIndex].buttonBindings[self->gamepadMaps[gamepadIndex].buttonBindingCount].actionID = actionID;
			self->gamepadMaps[gamepadIndex].buttonBindings[self->gamepadMaps[gamepadIndex].buttonBindingCount].buttonID = buttonID;
			self->gamepadMaps[gamepadIndex].buttonBindingCount++;
			
			event.vendorID = vendorID;
			event.productID = productID;
			event.binding = self->gamepadMaps[gamepadIndex].buttonBindings[self->gamepadMaps[gamepadIndex].buttonBindingCount - 1];
			self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_BUTTON_BOUND), &event);
			
			return;
		}
	}
	self->gamepadMaps = realloc(self->gamepadMaps, sizeof(struct PlayerInputMap_gamepadMap) * (self->gamepadMapCount + 1));
	self->gamepadMaps[self->gamepadMapCount].vendorID = vendorID;
	self->gamepadMaps[self->gamepadMapCount].productID = productID;
	self->gamepadMaps[self->gamepadMapCount].buttonBindingCount = 1;
	self->gamepadMaps[self->gamepadMapCount].buttonBindings = malloc(sizeof(struct PlayerInputMap_gamepadButtonBinding));
	self->gamepadMaps[self->gamepadMapCount].buttonBindings[0].actionID = actionID;
	self->gamepadMaps[self->gamepadMapCount].buttonBindings[0].buttonID = buttonID;
	self->gamepadMaps[self->gamepadMapCount].axisBindingCount = 0;
	self->gamepadMaps[self->gamepadMapCount].axisBindings = NULL;
	self->gamepadMapCount++;
	
	event.vendorID = vendorID;
	event.productID = productID;
	event.binding = self->gamepadMaps[self->gamepadMapCount - 1].buttonBindings[0];
	self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_BUTTON_BOUND), &event);
}

void PlayerInputMap_unbindButton(compat_type(PlayerInputMap *) selfPtr, Atom actionID, int vendorID, int productID, unsigned int buttonID) {
	PlayerInputMap * self = selfPtr;
	size_t gamepadIndex, bindingIndex;
	struct PlayerInputMap_buttonBindingEvent event;
	
	for (gamepadIndex = 0; gamepadIndex < self->gamepadMapCount; gamepadIndex++) {
		if (self->gamepadMaps[gamepadIndex].vendorID == vendorID && self->gamepadMaps[gamepadIndex].productID == productID) {
			for (bindingIndex = 0; bindingIndex < self->gamepadMaps[gamepadIndex].buttonBindingCount; bindingIndex++) {
				if (self->gamepadMaps[gamepadIndex].buttonBindings[bindingIndex].actionID == actionID && self->gamepadMaps[gamepadIndex].buttonBindings[bindingIndex].buttonID == buttonID) {
					event.vendorID = vendorID;
					event.productID = productID;
					event.binding = self->gamepadMaps[gamepadIndex].buttonBindings[bindingIndex];
					self->gamepadMaps[gamepadIndex].buttonBindingCount--;
					for (; bindingIndex < self->gamepadMaps[gamepadIndex].buttonBindingCount; bindingIndex++) {
						self->gamepadMaps[gamepadIndex].buttonBindings[bindingIndex] = self->gamepadMaps[gamepadIndex].buttonBindings[bindingIndex + 1];
					}
					self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_BUTTON_UNBOUND), &event);
					return;
				}
			}
		}
	}
}

bool PlayerInputMap_isAxisBound(compat_type(PlayerInputMap *) selfPtr, Atom actionID, int vendorID, int productID, unsigned int axisID, int direction) {
	PlayerInputMap * self = selfPtr;
	size_t gamepadIndex, bindingIndex;
	
	for (gamepadIndex = 0; gamepadIndex < self->gamepadMapCount; gamepadIndex++) {
		if (self->gamepadMaps[gamepadIndex].vendorID == vendorID && self->gamepadMaps[gamepadIndex].productID == productID) {
			for (bindingIndex = 0; bindingIndex < self->gamepadMaps[gamepadIndex].axisBindingCount; bindingIndex++) {
				if (self->gamepadMaps[gamepadIndex].axisBindings[bindingIndex].actionID == actionID && self->gamepadMaps[gamepadIndex].axisBindings[bindingIndex].axisID == axisID && self->gamepadMaps[gamepadIndex].axisBindings[bindingIndex].direction == direction) {
					return true;
				}
			}
		}
	}
	return false;
}

void PlayerInputMap_bindAxis(compat_type(PlayerInputMap *) selfPtr, Atom actionID, int vendorID, int productID, unsigned int axisID, int direction, float triggerThreshold, float releaseThreshold) {
	PlayerInputMap * self = selfPtr;
	size_t gamepadIndex, bindingIndex;
	struct PlayerInputMap_axisBindingEvent event;
	
	for (gamepadIndex = 0; gamepadIndex < self->gamepadMapCount; gamepadIndex++) {
		if (self->gamepadMaps[gamepadIndex].vendorID == vendorID && self->gamepadMaps[gamepadIndex].productID == productID) {
			for (bindingIndex = 0; bindingIndex < self->gamepadMaps[gamepadIndex].axisBindingCount; bindingIndex++) {
				if (self->gamepadMaps[gamepadIndex].axisBindings[bindingIndex].actionID == actionID && self->gamepadMaps[gamepadIndex].axisBindings[bindingIndex].axisID == axisID && self->gamepadMaps[gamepadIndex].axisBindings[bindingIndex].direction == direction) {
					self->gamepadMaps[gamepadIndex].axisBindings[bindingIndex].triggerThreshold = triggerThreshold;
					self->gamepadMaps[gamepadIndex].axisBindings[bindingIndex].releaseThreshold = releaseThreshold;
					return;
				}
			}
			self->gamepadMaps[gamepadIndex].axisBindings = realloc(self->gamepadMaps[gamepadIndex].axisBindings, sizeof(struct PlayerInputMap_gamepadAxisBinding) * (self->gamepadMaps[gamepadIndex].axisBindingCount + 1));
			self->gamepadMaps[gamepadIndex].axisBindings[self->gamepadMaps[gamepadIndex].axisBindingCount].actionID = actionID;
			self->gamepadMaps[gamepadIndex].axisBindings[self->gamepadMaps[gamepadIndex].axisBindingCount].axisID = axisID;
			self->gamepadMaps[gamepadIndex].axisBindings[self->gamepadMaps[gamepadIndex].axisBindingCount].direction = direction;
			self->gamepadMaps[gamepadIndex].axisBindings[self->gamepadMaps[gamepadIndex].axisBindingCount].triggerThreshold = triggerThreshold;
			self->gamepadMaps[gamepadIndex].axisBindings[self->gamepadMaps[gamepadIndex].axisBindingCount].releaseThreshold = releaseThreshold;
			self->gamepadMaps[gamepadIndex].axisBindingCount++;
			
			event.vendorID = vendorID;
			event.productID = productID;
			event.binding = self->gamepadMaps[gamepadIndex].axisBindings[self->gamepadMaps[gamepadIndex].axisBindingCount - 1];
			self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_AXIS_BOUND), &event);
			
			return;
		}
	}
	self->gamepadMaps = realloc(self->gamepadMaps, sizeof(struct PlayerInputMap_gamepadMap) * (self->gamepadMapCount + 1));
	self->gamepadMaps[self->gamepadMapCount].vendorID = vendorID;
	self->gamepadMaps[self->gamepadMapCount].productID = productID;
	self->gamepadMaps[self->gamepadMapCount].buttonBindingCount = 0;
	self->gamepadMaps[self->gamepadMapCount].buttonBindings = NULL;
	self->gamepadMaps[self->gamepadMapCount].axisBindingCount = 1;
	self->gamepadMaps[self->gamepadMapCount].axisBindings = malloc(sizeof(struct PlayerInputMap_gamepadAxisBinding));
	self->gamepadMaps[self->gamepadMapCount].axisBindings[0].actionID = actionID;
	self->gamepadMaps[self->gamepadMapCount].axisBindings[0].axisID = axisID;
	self->gamepadMaps[self->gamepadMapCount].axisBindings[0].direction = direction;
	self->gamepadMaps[self->gamepadMapCount].axisBindings[0].triggerThreshold = triggerThreshold;
	self->gamepadMaps[self->gamepadMapCount].axisBindings[0].releaseThreshold = releaseThreshold;
	self->gamepadMapCount++;
	
	event.vendorID = vendorID;
	event.productID = productID;
	event.binding = self->gamepadMaps[self->gamepadMapCount - 1].axisBindings[0];
	self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_AXIS_BOUND), &event);
}

void PlayerInputMap_unbindAxis(compat_type(PlayerInputMap *) selfPtr, Atom actionID, int vendorID, int productID, unsigned int axisID, int direction) {
	PlayerInputMap * self = selfPtr;
	size_t gamepadIndex, bindingIndex;
	struct PlayerInputMap_axisBindingEvent event;
	
	for (gamepadIndex = 0; gamepadIndex < self->gamepadMapCount; gamepadIndex++) {
		if (self->gamepadMaps[gamepadIndex].vendorID == vendorID && self->gamepadMaps[gamepadIndex].productID == productID) {
			for (bindingIndex = 0; bindingIndex < self->gamepadMaps[gamepadIndex].axisBindingCount; bindingIndex++) {
				if (self->gamepadMaps[gamepadIndex].axisBindings[bindingIndex].actionID == actionID && self->gamepadMaps[gamepadIndex].axisBindings[bindingIndex].axisID == axisID && self->gamepadMaps[gamepadIndex].axisBindings[bindingIndex].direction == direction) {
					event.vendorID = vendorID;
					event.productID = productID;
					event.binding = self->gamepadMaps[gamepadIndex].axisBindings[bindingIndex];
					self->gamepadMaps[gamepadIndex].axisBindingCount--;
					for (; bindingIndex < self->gamepadMaps[gamepadIndex].axisBindingCount; bindingIndex++) {
						self->gamepadMaps[gamepadIndex].axisBindings[bindingIndex] = self->gamepadMaps[gamepadIndex].axisBindings[bindingIndex + 1];
					}
					self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_INPUT_MAP_EVENT_AXIS_UNBOUND), &event);
					return;
				}
			}
			return;
		}
	}
}
