/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#ifndef __PLAYER_INPUT_MAP_H__
#define __PLAYER_INPUT_MAP_H__

typedef struct PlayerInputMap PlayerInputMap;

#include "serialization/DeserializationContext.h"
#include "serialization/SerializationContext.h"
#include "stemobject/StemObject.h"
#include "utilities/Atom.h"
#include "utilities/EventDispatcher.h"
#include <stdlib.h>

#define PLAYER_INPUT_MAP_SERIALIZATION_FORMAT_VERSION 2

// struct PlayerInputMap_keyboardBinding * eventData
#define PLAYER_INPUT_MAP_EVENT_KEY_BOUND "PLAYER_INPUT_MAP_EVENT_KEY_BOUND"
#define PLAYER_INPUT_MAP_EVENT_KEY_UNBOUND "PLAYER_INPUT_MAP_EVENT_KEY_UNBOUND"
// struct PlayerInputMap_buttonBindingEvent * eventData
#define PLAYER_INPUT_MAP_EVENT_BUTTON_BOUND "PLAYER_INPUT_MAP_EVENT_BUTTON_BOUND"
#define PLAYER_INPUT_MAP_EVENT_BUTTON_UNBOUND "PLAYER_INPUT_MAP_EVENT_BUTTON_UNBOUND"
// struct PlayerInputMap_axisBindingEvent * eventData
#define PLAYER_INPUT_MAP_EVENT_AXIS_BOUND "PLAYER_INPUT_MAP_EVENT_AXIS_BOUND"
#define PLAYER_INPUT_MAP_EVENT_AXIS_UNBOUND "PLAYER_INPUT_MAP_EVENT_AXIS_UNBOUND"

struct PlayerInputMap_keyboardBinding {
	Atom actionID;
	unsigned int keyCode;
	unsigned int charCode;
};

struct PlayerInputMap_gamepadButtonBinding {
	Atom actionID;
	unsigned int buttonID;
};

struct PlayerInputMap_gamepadAxisBinding {
	Atom actionID;
	unsigned int axisID;
	int direction;
	float triggerThreshold;
	float releaseThreshold;
};

struct PlayerInputMap_buttonBindingEvent {
	int vendorID;
	int productID;
	struct PlayerInputMap_gamepadButtonBinding binding;
};

struct PlayerInputMap_axisBindingEvent {
	int vendorID;
	int productID;
	struct PlayerInputMap_gamepadAxisBinding binding;
};

struct PlayerInputMap_gamepadMap {
	int vendorID;
	int productID;
	size_t buttonBindingCount;
	struct PlayerInputMap_gamepadButtonBinding * buttonBindings;
	size_t axisBindingCount;
	struct PlayerInputMap_gamepadAxisBinding * axisBindings;
};

#define PlayerInputMap_structContents \
	StemObject_structContents \
	\
	EventDispatcher * eventDispatcher; \
	size_t keyboardBindingCount; \
	struct PlayerInputMap_keyboardBinding * keyboardBindings; \
	size_t gamepadMapCount; \
	struct PlayerInputMap_gamepadMap * gamepadMaps; \
	\
	bool (* isKeyBound)(compat_type(PlayerInputMap *) self, Atom actionID, unsigned int keyCode); \
	void (* bindKey)(compat_type(PlayerInputMap *) self, Atom actionID, unsigned int charCode, unsigned int keyCode); \
	void (* unbindKey)(compat_type(PlayerInputMap *) self, Atom actionID, unsigned int keyCode); \
	bool (* isButtonBound)(compat_type(PlayerInputMap *) self, Atom actionID, int vendorID, int productID, unsigned int buttonID); \
	void (* bindButton)(compat_type(PlayerInputMap *) self, Atom actionID, int vendorID, int productID, unsigned int buttonID); \
	void (* unbindButton)(compat_type(PlayerInputMap *) self, Atom actionID, int vendorID, int productID, unsigned int buttonID); \
	bool (* isAxisBound)(compat_type(PlayerInputMap *) self, Atom actionID, int vendorID, int productID, unsigned int axisID, int direction); \
	void (* bindAxis)(compat_type(PlayerInputMap *) self, Atom actionID, int vendorID, int productID, unsigned int axisID, int direction, float triggerThreshold, float releaseThreshold); \
	void (* unbindAxis)(compat_type(PlayerInputMap *) self, Atom actionID, int vendorID, int productID, unsigned int axisID, int direction);

struct PlayerInputMap {
	PlayerInputMap_structContents
};

PlayerInputMap * PlayerInputMap_create();
void PlayerInputMap_init(compat_type(PlayerInputMap *) selfPtr);
void PlayerInputMap_dispose(compat_type(PlayerInputMap *) selfPtr);

PlayerInputMap * PlayerInputMap_deserialize(compat_type(DeserializationContext *) deserializationContext);
bool PlayerInputMap_loadSerializedData(compat_type(PlayerInputMap *) selfPtr, compat_type(DeserializationContext *) deserializationContext);
void PlayerInputMap_serialize(compat_type(PlayerInputMap *) selfPtr, compat_type(SerializationContext *) serializationContext);

bool PlayerInputMap_isKeyBound(compat_type(PlayerInputMap *) selfPtr, Atom actionID, unsigned int keyCode);
void PlayerInputMap_bindKey(compat_type(PlayerInputMap *) selfPtr, Atom actionID, unsigned int charCode, unsigned int keyCode);
void PlayerInputMap_unbindKey(compat_type(PlayerInputMap *) selfPtr, Atom actionID, unsigned int keyCode);
bool PlayerInputMap_isButtonBound(compat_type(PlayerInputMap *) selfPtr, Atom actionID, int vendorID, int productID, unsigned int buttonID);
void PlayerInputMap_bindButton(compat_type(PlayerInputMap *) selfPtr, Atom actionID, int vendorID, int productID, unsigned int buttonID);
void PlayerInputMap_unbindButton(compat_type(PlayerInputMap *) selfPtr, Atom actionID, int vendorID, int productID, unsigned int buttonID);
bool PlayerInputMap_isAxisBound(compat_type(PlayerInputMap *) selfPtr, Atom actionID, int vendorID, int productID, unsigned int axisID, int direction);
void PlayerInputMap_bindAxis(compat_type(PlayerInputMap *) selfPtr, Atom actionID, int vendorID, int productID, unsigned int axisID, int direction, float triggerThreshold, float releaseThreshold);
void PlayerInputMap_unbindAxis(compat_type(PlayerInputMap *) selfPtr, Atom actionID, int vendorID, int productID, unsigned int axisID, int direction);

#endif
