/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/RecordingPlayerInputController.h"
#include "jsonserialization/JSONSerializationContext.h"
#include <stdio.h>

RecordingPlayerInputController * RecordingPlayerInputController_create(PlayerModel * playerModel, PlayerInputMap * inputMap) {
	stemobject_create_implementation(RecordingPlayerInputController, init, playerModel, inputMap)
}

void RecordingPlayerInputController_init(compat_type(RecordingPlayerInputController *) selfPtr, PlayerModel * playerModel, PlayerInputMap * inputMap) {
	RecordingPlayerInputController * self = selfPtr;
	
	PlayerInputController_init(self, playerModel, inputMap);
	self->dispose = RecordingPlayerInputController_dispose;
	self->protected_ivar(triggerAction) = RecordingPlayerInputController_triggerAction;
	self->protected_ivar(releaseAction) = RecordingPlayerInputController_releaseAction;
	self->update = RecordingPlayerInputController_update;
	self->save = RecordingPlayerInputController_save;
	
	self->frameCount = 0;
	self->nodeCount = 0;
	self->nodeListSize = 16;
	self->nodes = malloc(sizeof(struct RecordedPlayerInputNode) * self->nodeListSize);
}

void RecordingPlayerInputController_dispose(compat_type(RecordingPlayerInputController *) selfPtr) {
	RecordingPlayerInputController * self = selfPtr;
	
	free(self->nodes);
	PlayerInputController_dispose(self);
}

static void addNode(RecordingPlayerInputController * self, struct RecordedPlayerInputNode node) {
	if (self->nodeCount >= self->nodeListSize) {
		self->nodeListSize *= 2;
		self->nodes = realloc(self->nodes, sizeof(struct RecordedPlayerInputNode) * self->nodeListSize);
	}
	self->nodes[self->nodeCount++] = node;
}

bool RecordingPlayerInputController_triggerAction(compat_type(RecordingPlayerInputController *) selfPtr, Atom actionID) {
	RecordingPlayerInputController * self = selfPtr;
	bool apply = false;
	
	if (actionID == Atom_fromString("left")) {
		apply = !self->leftTriggered;
	} else if (actionID == Atom_fromString("right")) {
		apply = !self->rightTriggered;
	} else if (actionID == Atom_fromString("down")) {
		apply = !self->downTriggered;
	} else if (actionID == Atom_fromString("up")) {
		apply = !self->upTriggered;
	} else if (actionID == Atom_fromString("attack")) {
		apply = !self->attackTriggered;
	} else if (actionID == Atom_fromString("special")) {
		apply = !self->specialTriggered;
	}
	
	if (apply) {
		addNode(self, (struct RecordedPlayerInputNode) {self->frameCount, actionID, true});
	}
	
	return PlayerInputController_triggerAction(self, actionID);
}

bool RecordingPlayerInputController_releaseAction(compat_type(RecordingPlayerInputController *) selfPtr, Atom actionID) {
	RecordingPlayerInputController * self = selfPtr;
	bool apply = false;
	
	if (actionID == Atom_fromString("left")) {
		apply = self->leftTriggered;
	} else if (actionID == Atom_fromString("right")) {
		apply = self->rightTriggered;
	} else if (actionID == Atom_fromString("down")) {
		apply = self->downTriggered;
	} else if (actionID == Atom_fromString("up")) {
		apply = self->upTriggered;
	} else if (actionID == Atom_fromString("attack")) {
		apply = self->attackTriggered;
	} else if (actionID == Atom_fromString("special")) {
		apply = self->specialTriggered;
	}
	
	if (apply) {
		addNode(self, (struct RecordedPlayerInputNode) {self->frameCount, actionID, false});
	}
	
	return PlayerInputController_releaseAction(self, actionID);
}

void RecordingPlayerInputController_update(compat_type(RecordingPlayerInputController *) selfPtr) {
	RecordingPlayerInputController * self = selfPtr;
	
	self->frameCount++;
	InputController_update(self);
}

void RecordingPlayerInputController_save(compat_type(RecordingPlayerInputController *) selfPtr, const char * fileName) {
	RecordingPlayerInputController * self = selfPtr;
	JSONSerializationContext * context;
	size_t nodeIndex;
	
	context = JSONSerializationContext_create();
	context->beginStructure(context, "input_session");
	context->writeUInt16(context, "format_version", PLAYER_INPUT_SESSION_VERSION);
	context->beginArray(context, "events");
	for (nodeIndex = 0; nodeIndex < self->nodeCount; nodeIndex++) {
		context->beginStructure(context, NULL);
		context->writeUInt32(context, "frame", self->nodes[nodeIndex].frameIndex);
		context->writeString(context, "action", self->nodes[nodeIndex].actionID);
		context->writeBoolean(context, "triggered", self->nodes[nodeIndex].triggered);
		context->endStructure(context);
	}
	context->endArray(context);
	context->endStructure(context);
	
	if (!context->writeToFile(context, JSONEmitterFormat_multiLine, fileName, NULL)) {
		fprintf(stderr, "Error: Couldn't write \"%s\"; input session not saved\n", fileName);
	}
	context->dispose(context);
}
